/*
  The partysip program is a modular SIP proxy server (SIP -rfc3261-)
  Copyright (C) 2002  WellX Telecom   - <partysip@wellx.com>
  Copyright (C) 2002  Aymeric MOIZARD - <jack@atosc.org>
  
  Partysip is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  Partysip is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with Foobar; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef _PARTYSIP_H_
#define _PARTYSIP_H_

#ifndef WIN32
#include "config.h"
#endif

#include <ppl/ppl_time.h>
#include <ppl/ppl_dso.h>
#include <ppl/ppl_socket.h>
#include <ppl/ppl_pipe.h>

#include <osip/osip.h>
#include <osip/smsg.h>
#include <osip/thread.h>

#include <partysip/psp_macros.h>
#include <partysip/psp_req.h>
#include <partysip/psp_plug.h>
#include <partysip/psp_config.h>

#include <partysip/psp_nat.h>

#ifndef INVITE_TIMEOUT
#define INVITE_TIMEOUT   (60)
#endif

/* hidden (from user-point of view) config */
typedef struct _psp_osip_t
{
  osip_t *osip;			/* osip main element                         */
  fifo_t *osip_tr_terminated;	/* osip transactions in the terminated state */
  int delay;			/* delay between update (should be 500ms)    */
  int exit_flag;		/* quick hack to shutdown the timers thread  */
  sthread_t *timers;		/* internal osip timer handler               */

  ppl_pipe_t *wakeup;

  /* basic statistics */
  int total_ict;
  int successfull_ict;
  int total_nict;
  int successfull_nict;
  int total_ist;
  int successfull_ist;
  int total_nist;
  int successfull_nist;

}
psp_osip_t;

int psp_osip_init (psp_osip_t ** psp_osip, int delay);
void psp_osip_free (psp_osip_t * psp_osip);
int psp_osip_wakeup (psp_osip_t * psp_osip);
int psp_osip_release (psp_osip_t * psp_osip);
int psp_osip_set_delay (psp_osip_t * psp_osip, int delay);


typedef struct _module_t
{

#define MOD_THREAD     1	/* start module_execute() once */
  int flag;			/* MOD_THREAD */
  sthread_t *thread;		/* modules are initiated in a thread (optionnaly) */
  ppl_pipe_t *wakeup;		/* descriptor used to wake up the thread */
  char *filename;
}
module_t;

int module_init (module_t ** module, int flag, char *filename);
void module_free (module_t * module);
int module_release (module_t * module);
int module_start (module_t * module, void *(*func_start) (void *), void *arg);
int module_set_flag (module_t * module, int flag);
int module_wakeup (module_t * module);

/********************************************************************/
/**************************   resolver    ***************************/

typedef struct _pspm_resolv_t
{
  module_t *module;

  /* no plugin */
}
pspm_resolv_t;

int pspm_resolv_init (pspm_resolv_t ** resolv);
void pspm_resolv_free (pspm_resolv_t * resolv);
int pspm_resolv_release (pspm_resolv_t * resolv);
int pspm_resolv_start (pspm_resolv_t * resolv, void *(*func_start) (void *),
		       void *arg);
int pspm_resolv_execute (pspm_resolv_t * resolv, int sec_max, int usec_max,
			 int max);
void pspm_resolv_handle_dns_query (pspm_resolv_t * resolv, char *address);


/********************************************************************/
/*****************************   tlp    *****************************/

typedef struct _pspm_tlp_t
{
  module_t *module;

  tlp_plugin_t *tlp_plugins;	/* list of tlp_plugins */

}
pspm_tlp_t;

int pspm_tlp_init (pspm_tlp_t ** pspm);
void pspm_tlp_free (pspm_tlp_t * pspm);
int pspm_tlp_execute (pspm_tlp_t * pspm, int sec_max,
		      int usec_max, int max_analysed);
int pspm_tlp_start (pspm_tlp_t * pspm, void *(*func_start) (void *),
		    void *arg);
int pspm_tlp_release (pspm_tlp_t * pspm);

/* int  pspm_tlp_add_plugin(pspm_tlp_t *m, psp_plugin_t* plugin) */
int pspm_tlp_send_message (pspm_tlp_t * pspm, transaction_t * tr,
			   sip_t * sip, char *host, int port, int out_socket);

/********************************************************************/
/*****************************   imp   *****************************/

/* this modules handles all sip_t incoming request.
   It first set up a psp_req_t context
   Call plugins' registred methods
   Dispatch the result to the correct module
*/
typedef struct _pspm_imp_t
{
  module_t *module;

  imp_plugin_t *imp_plugins;	/* order does matter -> up to the administrator */

  imp_func_tab_t *func_invites;	/* reception of INVITE    */
  imp_func_tab_t *func_acks;	/* reception of ACK       */
  imp_func_tab_t *func_registers;	/*              REGISTER  */
  imp_func_tab_t *func_byes;	/*              BYE       */
  imp_func_tab_t *func_optionss;	/*              OPTIONS   */
  imp_func_tab_t *func_infos;	/*              INFO      */
  imp_func_tab_t *func_cancels;	/*              CANCEL    */
  imp_func_tab_t *func_notifys;	/*              NOTIFY    */
  imp_func_tab_t *func_subscribes;	/*              SUBSCRIBE */
  imp_func_tab_t *func_unknowns;	/*              UNKNOWN   */

  fifo_t *sip_ack;		/* sip_t ACK messages */
  fifo_t *sip_traffic;		/* New transaction_t element. */
}
pspm_imp_t;

int pspm_imp_init (pspm_imp_t ** pspm);
void pspm_imp_free (pspm_imp_t * pspm);
int pspm_imp_execute (pspm_imp_t * pspm, int sec_max, int usec_max, int max);
int pspm_imp_release (pspm_imp_t * pspm);
int pspm_imp_start (pspm_imp_t * pspm, void *(*func_start) (void *),
		    void *arg);


int pspm_imp_dispatch_ack (pspm_imp_t * pspm, sip_t * ack);
int pspm_imp_dispatch_traffic (pspm_imp_t * pspm, transaction_t * inc_tr);
int pspm_imp_call_plugins (pspm_imp_t * pspm, psp_req_t * req);
int pspm_imp_dispatch_psp_request (pspm_imp_t * pspm, psp_req_t * req);

/********************************************************************/
/*****************************   uap   ******************************/

typedef struct _pspm_uap_t
{
  /* internal use only */
  module_t *module;

  uap_plugin_t *uap_plugins;

  uap_func_tab_t *func_1xxs;	/* sending      1xx    */
  uap_func_tab_t *func_2xxs;	/*              2xx    */
  uap_func_tab_t *func_3xxs;	/*              3xx    */
  uap_func_tab_t *func_4xxs;	/*              4xx    */
  uap_func_tab_t *func_5xxs;	/*              5xx    */
  uap_func_tab_t *func_6xxs;	/*              6xx    */

  fifo_t *uas_request;		/* psp_req_t ELEMENTs */
  psp_req_t *uap_requests;	/* psp_req_t ELEMENTs */
}
pspm_uap_t;

int pspm_uap_init (pspm_uap_t ** pspm);
void pspm_uap_free (pspm_uap_t * pspm);
int pspm_uap_execute (pspm_uap_t * pspm, int sec_max, int usec_max, int max);
int pspm_uap_release (pspm_uap_t * pspm);
int pspm_uap_start (pspm_uap_t * pspm, void *(*func_start) (void *),
		    void *arg);
int pspm_uap_handle_psp_request (pspm_uap_t * pspm, psp_req_t * req);
int pspm_uap_call_plugins (pspm_uap_t * pspm, psp_req_t * req);
int pspm_uap_delete_terminated_transaction (pspm_uap_t * pspm,
					    psp_req_t * req);
int pspm_uap_delete_broken_transaction (pspm_uap_t * pspm, psp_req_t * req);

/********************************************************************/
/*****************************   sfp   *****************************/

typedef struct _pspm_sfp_t
{
  /* internal use only */
  module_t *module;

  sfp_plugin_t *sfp_plugins;

  sfp_fwd_func_tab_t *fwd_invites;	/* forward      INVITE    */
  sfp_fwd_func_tab_t *fwd_acks;	/* forward      ACK       */
  sfp_fwd_func_tab_t *fwd_registers;	/*              REGISTER  */
  sfp_fwd_func_tab_t *fwd_byes;	/*              BYE       */
  sfp_fwd_func_tab_t *fwd_optionss;	/*              OPTIONS   */
  sfp_fwd_func_tab_t *fwd_infos;	/*              INFO      */
  sfp_fwd_func_tab_t *fwd_cancels;	/*              CANCEL    */
  sfp_fwd_func_tab_t *fwd_notifys;	/*              NOTIFY    */
  sfp_fwd_func_tab_t *fwd_subscribes;	/*              SUBSCRIBE */
  sfp_fwd_func_tab_t *fwd_unknowns;	/*              UNKNOWN   */

  sfp_rcv_func_tab_t *rcv_1xxs;	/* receive      1xx    */
  sfp_rcv_func_tab_t *rcv_2xxs;	/*              2xx    */
  sfp_rcv_func_tab_t *rcv_3xxs;	/*              3xx    */
  sfp_rcv_func_tab_t *rcv_4xxs;	/*              4xx    */
  sfp_rcv_func_tab_t *rcv_5xxs;	/*              5xx    */
  sfp_rcv_func_tab_t *rcv_6xxs;	/*              6xx    */

  sfp_snd_func_tab_t *snd_1xxs;	/* forward      1xx    */
  sfp_snd_func_tab_t *snd_2xxs;	/*              2xx    */
  sfp_snd_func_tab_t *snd_3xxs;	/*              3xx    */
  sfp_snd_func_tab_t *snd_4xxs;	/*              4xx    */
  sfp_snd_func_tab_t *snd_5xxs;	/*              5xx    */
  sfp_snd_func_tab_t *snd_6xxs;	/*              6xx    */

  fifo_t *sfull_request;	/* psp_req_t ELEMENT in STATEFULL_MODE */
  fifo_t *sfull_cancels;	/* sip_t ELEMENT */

  list_t *broken_transactions;
  /* Definitions of several STATEs for psp_req. Each defines the conditions
     required and the actions to be taken.

     1: waiting for new location to try
     CONDITIONS: req->locations->url->host is an ip address.
     or ppl_dns_get_result (&dns_result, loc->url->host); give a result.
     ACTION: create an outgoing transaction with the new location.
     2: waiting for informative answers in outgoing transactions
     CONDITIONS: psp_req->rsp_ctx->branches->transaction->state==XICT_PROCEEDING
     && psp_req->rsp_ctx->branches->last_status <
     atoi(psp_req->rsp_ctx->branches->transaction->last_response->strtline->statuscode)))
     ACTION:  discard status==100
     ACTION2: forward response.
     3: waiting for final answer:
     CONDITIONS: psp_req->rsp_ctx->branches->transaction->state
     ==XICT_COMPLETED||==XICT_TERMINATED||
     && 199 <
     atoi(psp_req->rsp_ctx->branches->transaction->last_response->strtline->statuscode)))
     ACTION: if response is 200 ==> forward response AND cancel other branches.
     ACTION2: if response is >199 AND no other branch => forward response.
     3: waiting for transaction timeout:
     CONDITIONS: psp_req->rsp_ctx->branches->transaction->state==XICT_PROCEEDING
     && psp_req->rsp_ctx->branches is expired
     ACTION: if response is 100 mark branch as deleted.
     ACTION2: if response is 101 to 199, cancel branch.
   */
  /*  smutex_t *mut_psp_requests; */
  psp_req_t *psp_requests;
}
pspm_sfp_t;


int pspm_sfp_init (pspm_sfp_t ** pspm);
void pspm_sfp_free (pspm_sfp_t * pspm);
int pspm_sfp_execute (pspm_sfp_t * pspm, int sec_max, int usec_max, int max);
int pspm_sfp_start (pspm_sfp_t * pspm, void *(*func_start) (void *),
		    void *arg);
int pspm_sfp_release (pspm_sfp_t * pspm);

/*
  int pspm_sfp_lock_psp_requests_access (pspm_sfp_t * pspm);
  int pspm_sfp_unlock_psp_requests_access (pspm_sfp_t * pspm); */

int pspm_sfp_handle_new_cancel (pspm_sfp_t * pspm, transaction_t * tr_cancel);
int pspm_sfp_handle_new_request (pspm_sfp_t * pspm, psp_req_t * req);
int pspm_sfp_handle_new_location (pspm_sfp_t * pspm, psp_req_t * req);
int pspm_sfp_delete_broken_location (pspm_sfp_t * pspm, psp_req_t * req);
int pspm_sfp_delete_completed_branchs (pspm_sfp_t * pspm, psp_req_t * req);

int pspm_all_branch_are_terminated (pspm_sfp_t * pspm, psp_req_t * req);
int pspm_sfp_forward_200ok_final_response (pspm_sfp_t * pspm,
					   psp_req_t * req);
int pspm_sfp_forward_provisionnal_response (pspm_sfp_t * pspm,
					    psp_req_t * req);
int pspm_all_request_are_terminated (pspm_sfp_t * pspm, psp_req_t * req);
int pspm_sfp_wait_for_final_response (pspm_sfp_t * pspm, psp_req_t * req);
int pspm_sfp_answer_request_and_cancel_branches (pspm_sfp_t * pspm,
						 psp_req_t * req, int code);

int pspm_sfp_cancel_all_pending_branches (pspm_sfp_t * pspm, psp_req_t * req);
int pspm_sfp_delete_broken_transactions (pspm_sfp_t * pspm);

/* int pspm_sfp_reply_best_answer_or_404 (pspm_sfp_t * pspm, psp_req_t * req); */


int pspm_sfp_call_plugins_for_request (pspm_sfp_t * pspm, psp_req_t * req,
				       sfp_branch_t * sfp_branch);
int pspm_sfp_call_plugins_for_rcv_response (pspm_sfp_t * pspm,
					    psp_req_t * req,
					    sfp_branch_t * branch);
int pspm_sfp_call_plugins_for_snd_response (pspm_sfp_t * pspm,
					    psp_req_t * req);

#ifndef MAX_SERVER_IDENTIFIERS
#define MAX_SERVER_IDENTIFIERS	32
#endif

typedef struct _psp_core_t
{

  psp_osip_t *psp_osip;		/* osip related params */

  /* core elements */
  pspm_resolv_t *resolv;	/* module for Address Resolutio (no plugin) */
  pspm_tlp_t *tlp;		/* module for Transport layer Processing:   */
  pspm_imp_t *imp;		/* module for Incoming Message Processing:  */
  pspm_uap_t *uap;		/* module for Ua Server Processing          */
  pspm_sfp_t *sfp;		/* module for State Full Processing         */

  int nat_enabled;		/* NAT feature enabled if == 1 */
  char *lan_ip;
  char *lan_fqdn;
  char *lan_mask;
  char *ext_ip;
  char *ext_fqdn;
  char *ext_mask;

  int iptsrv_port;		/* port for iptable_srv(a server clone if 'iptable') */
  char *iptsrv_srvip;		/* ip   for iptable_srv(a server clone if 'iptable') */

  char *serverip[MAX_SERVER_IDENTIFIERS];	/* IP addresses for this proxy */
  char *servername[MAX_SERVER_IDENTIFIERS];	/* Hostnames for this proxy
						 * (should be ALL valid hostnames)
						 */
  int port;			/* this proxy runs on this port.             */
  smutex_t *gl_lock;

  firewall_entry_t fw_entries[PSP_NAT_MAX_CALL];

  /* rfc3337.txt
     Extension Header Field for Registering Non-Adjacent Contacts */
  int rfc3327; /* if set: add a Path header in outgoing REGISTER messages */

  /* compatibility issues */
  /* this is for compatibility with some
     Uncompliant UAs (ATA v2.15 ata18x (020927a)) */
  int disable_check_for_to_tag_in_cancel;	/* 0 inactive, 1 active */


}
psp_core_t;

psp_core_t *psp_core_get ();
int psp_core_init ();
void psp_core_free ();

int psp_core_lock ();
int psp_core_unlock ();
int psp_core_start (int use_this_thread);
int psp_core_execute ();
void psp_core_remove_all_hook ();
void psp_core_remove_all_plugins ();
PPL_DECLARE (char *)
psp_core_get_ip_of_local_proxy ();
PPL_DECLARE (int)
psp_core_get_port_of_local_proxy ();
PPL_DECLARE (int)
psp_core_is_responsible_for_this_domain (url_t * url);
PPL_DECLARE (int)
psp_core_is_responsible_for_this_request_uri (url_t * url);
PPL_DECLARE (int)
psp_core_is_responsible_for_this_route (url_t * url);
     int psp_core_sfp_new_rsp_ctx (rsp_ctx_t ** ctx, psp_req_t * req);
     char *psp_core_sfp_generate_branch_for_request (sip_t * request);
     char *psp_core_default_generate_branch_for_request (sip_t * request);
     psp_req_t *psp_core_sfp_find_req (int osip_id);
PPL_DECLARE (int)
psp_core_find_transaction_and_add_event (sipevent_t * evt);
     transaction_t *psp_core_uap_create_transaction (sipevent_t * evt);
     int psp_core_switch_to_uas_mode (psp_req_t * req, int status);
PPL_DECLARE (int)
psp_core_fix_last_via_header (sipevent_t * evt, char *ip_addr, int port);
PPL_DECLARE (int)
psp_core_fix_strict_router_issue (sipevent_t * evt);

/* this method can handle late answer without calling plugins, this
   could lead to a response forwarded that does not match the
   previous ones. (in the case, some plugins act as if partysip
   is a B2BUA and for example modify the SDP body.) */
     int psp_core_handle_late_answer (sip_t * answer);

/* NETWORK related callbacks for osip */
/* Note: out_socket is not always set! For example, with UDP, we could
   always use the same global socket to send message. For TCP, we
   should create a new socket when a new transaction is created and
   this is the correct place to store the socket. In this case, you
   must also take care of closing the socket... */
     int psp_core_cb_snd_message (transaction_t * tr, sip_t * sip, char *host,
				  int port, int out_socket);

/* UAS related callbacks for osip */
     void psp_core_cb_endoftransaction (transaction_t * tr);
     void psp_core_cb_ist_kill_transaction (transaction_t * tr);
     void psp_core_cb_nist_kill_transaction (transaction_t * tr);
     void psp_core_cb_sndresp_retransmission (transaction_t * tr,
					      sip_t * sip);
     void psp_core_cb_rcvreq_retransmission (transaction_t * tr, sip_t * sip);
     void psp_core_cb_rcvinvite (transaction_t * tr, sip_t * sip);
     void psp_core_cb_rcvack (transaction_t * tr, sip_t * sip);
     void psp_core_cb_rcvack2 (transaction_t * tr, sip_t * sip);
     void psp_core_cb_rcvregister (transaction_t * tr, sip_t * sip);
     void psp_core_cb_rcvbye (transaction_t * tr, sip_t * sip);
     void psp_core_cb_rcvcancel (transaction_t * tr, sip_t * sip);
     void psp_core_cb_rcvinfo (transaction_t * tr, sip_t * sip);
     void psp_core_cb_rcvoptions (transaction_t * tr, sip_t * sip);
     void psp_core_cb_rcvprack (transaction_t * tr, sip_t * sip);
     void psp_core_cb_rcvunkrequest (transaction_t * tr, sip_t * sip);
     void psp_core_cb_snd1xx (transaction_t * tr, sip_t * sip);
     void psp_core_cb_snd2xx (transaction_t * tr, sip_t * sip);
     void psp_core_cb_snd3xx (transaction_t * tr, sip_t * sip);
     void psp_core_cb_snd4xx (transaction_t * tr, sip_t * sip);
     void psp_core_cb_snd5xx (transaction_t * tr, sip_t * sip);
     void psp_core_cb_snd6xx (transaction_t * tr, sip_t * sip);

/* UAC related callbacks for osip */
     void psp_core_cb_ict_kill_transaction (transaction_t * tr);
     void psp_core_cb_nict_kill_transaction (transaction_t * tr);
     void psp_core_cb_transport_error (transaction_t * tr, int error);
     void psp_core_cb_sndreq_retransmission (transaction_t * tr, sip_t * sip);
     void psp_core_cb_rcvresp_retransmission (transaction_t * tr,
					      sip_t * sip);
     void psp_core_cb_sndinvite (transaction_t * tr, sip_t * sip);
     void psp_core_cb_sndack (transaction_t * tr, sip_t * sip);
     void psp_core_cb_sndbye (transaction_t * tr, sip_t * sip);
     void psp_core_cb_sndcancel (transaction_t * tr, sip_t * sip);
     void psp_core_cb_sndinfo (transaction_t * tr, sip_t * sip);
     void psp_core_cb_sndoptions (transaction_t * tr, sip_t * sip);
     void psp_core_cb_sndregister (transaction_t * tr, sip_t * sip);
     void psp_core_cb_sndprack (transaction_t * tr, sip_t * sip);
     void psp_core_cb_sndunkrequest (transaction_t * tr, sip_t * sip);
     void psp_core_cb_rcv1xx (transaction_t * tr, sip_t * sip);
     void psp_core_cb_rcv2xx (transaction_t * tr, sip_t * sip);
     void psp_core_cb_rcv3xx (transaction_t * tr, sip_t * sip);
     void psp_core_cb_rcv4xx (transaction_t * tr, sip_t * sip);
     void psp_core_cb_rcv5xx (transaction_t * tr, sip_t * sip);
     void psp_core_cb_rcv6xx (transaction_t * tr, sip_t * sip);

/* this API must be available to plugins */
/* config step */
     int psp_core_load_psp_plugin (char *name, psp_plugin_t ** plug);

PPL_DECLARE (int)
psp_core_load_tlp_plugin (tlp_plugin_t ** p, psp_plugin_t * psp, int flag);
PPL_DECLARE (int)
psp_core_add_tlp_plugin (tlp_plugin_t * plug);

PPL_DECLARE (int)
psp_core_load_imp_plugin (imp_plugin_t ** p, psp_plugin_t * psp);
PPL_DECLARE (int)
psp_core_add_imp_plugin (imp_plugin_t * plug);

PPL_DECLARE (int)
psp_core_load_uap_plugin (uap_plugin_t ** p, psp_plugin_t * psp);
PPL_DECLARE (int)
psp_core_add_uap_plugin (uap_plugin_t * plug);

PPL_DECLARE (int)
psp_core_load_sfp_plugin (sfp_plugin_t ** p, psp_plugin_t * psp);
PPL_DECLARE (int)
psp_core_add_sfp_plugin (sfp_plugin_t * plug);

/***IMP***/
PPL_DECLARE (int)
psp_core_add_imp_invite_hook (imp_func_t * fn, int hookflg);
PPL_DECLARE (int)
psp_core_add_imp_ack_hook (imp_func_t * fn, int hookflg);
PPL_DECLARE (int)
psp_core_add_imp_register_hook (imp_func_t * fn, int hookflg);
PPL_DECLARE (int)
psp_core_add_imp_bye_hook (imp_func_t * fn, int hookflg);
PPL_DECLARE (int)
psp_core_add_imp_options_hook (imp_func_t * fn, int hookflg);
PPL_DECLARE (int)
psp_core_add_imp_info_hook (imp_func_t * fn, int hookflg);
PPL_DECLARE (int)
psp_core_add_imp_cancel_hook (imp_func_t * fn, int hookflg);
PPL_DECLARE (int)
psp_core_add_imp_notify_hook (imp_func_t * fn, int hookflg);
PPL_DECLARE (int)
psp_core_add_imp_subscribe_hook (imp_func_t * fn, int hookflg);
PPL_DECLARE (int)
psp_core_add_imp_unknown_hook (imp_func_t * fn, int hookflg);

/***UAP***/
PPL_DECLARE (int)
psp_core_add_uap_snd_1xx_hook (uap_func_t * fn, int hookflg);
PPL_DECLARE (int)
psp_core_add_uap_snd_2xx_hook (uap_func_t * fn, int hookflg);
PPL_DECLARE (int)
psp_core_add_uap_snd_3xx_hook (uap_func_t * fn, int hookflg);
PPL_DECLARE (int)
psp_core_add_uap_snd_4xx_hook (uap_func_t * fn, int hookflg);
PPL_DECLARE (int)
psp_core_add_uap_snd_5xx_hook (uap_func_t * fn, int hookflg);
PPL_DECLARE (int)
psp_core_add_uap_snd_6xx_hook (uap_func_t * fn, int hookflg);


/***SFP***/
PPL_DECLARE (int)
psp_core_add_sfp_fwd_invite_hook (sfp_fwd_func_t * fn, int hookflg);
PPL_DECLARE (int)
psp_core_add_sfp_fwd_ack_hook (sfp_fwd_func_t * fn, int hookflg);
PPL_DECLARE (int)
psp_core_add_sfp_fwd_register_hook (sfp_fwd_func_t * fn, int hookflg);
PPL_DECLARE (int)
psp_core_add_sfp_fwd_bye_hook (sfp_fwd_func_t * fn, int hookflg);
PPL_DECLARE (int)
psp_core_add_sfp_fwd_options_hook (sfp_fwd_func_t * fn, int hookflg);
PPL_DECLARE (int)
psp_core_add_sfp_fwd_info_hook (sfp_fwd_func_t * fn, int hookflg);
PPL_DECLARE (int)
psp_core_add_sfp_fwd_cancel_hook (sfp_fwd_func_t * fn, int hookflg);
PPL_DECLARE (int)
psp_core_add_sfp_fwd_notify_hook (sfp_fwd_func_t * fn, int hookflg);
PPL_DECLARE (int)
psp_core_add_sfp_fwd_subscribe_hook (sfp_fwd_func_t * fn, int hookflg);
PPL_DECLARE (int)
psp_core_add_sfp_fwd_unknown_hook (sfp_fwd_func_t * fn, int hookflg);

PPL_DECLARE (int)
psp_core_add_sfp_rcv_1xx_hook (sfp_rcv_func_t * fn, int hookflg);
PPL_DECLARE (int)
psp_core_add_sfp_rcv_2xx_hook (sfp_rcv_func_t * fn, int hookflg);
PPL_DECLARE (int)
psp_core_add_sfp_rcv_3xx_hook (sfp_rcv_func_t * fn, int hookflg);
PPL_DECLARE (int)
psp_core_add_sfp_rcv_4xx_hook (sfp_rcv_func_t * fn, int hookflg);
PPL_DECLARE (int)
psp_core_add_sfp_rcv_5xx_hook (sfp_rcv_func_t * fn, int hookflg);
PPL_DECLARE (int)
psp_core_add_sfp_rcv_6xx_hook (sfp_rcv_func_t * fn, int hookflg);

PPL_DECLARE (int)
psp_core_add_sfp_snd_1xx_hook (sfp_snd_func_t * fn, int hookflg);
PPL_DECLARE (int)
psp_core_add_sfp_snd_2xx_hook (sfp_snd_func_t * fn, int hookflg);
PPL_DECLARE (int)
psp_core_add_sfp_snd_3xx_hook (sfp_snd_func_t * fn, int hookflg);
PPL_DECLARE (int)
psp_core_add_sfp_snd_4xx_hook (sfp_snd_func_t * fn, int hookflg);
PPL_DECLARE (int)
psp_core_add_sfp_snd_5xx_hook (sfp_snd_func_t * fn, int hookflg);
PPL_DECLARE (int)
psp_core_add_sfp_snd_6xx_hook (sfp_snd_func_t * fn, int hookflg);

/* run time step */
PPL_DECLARE (int)
psp_core_dns_add_domain_error (char *address);
PPL_DECLARE (int)
psp_core_dns_add_domain_result (ppl_dns_entry_t * dns);
PPL_DECLARE (int)
psp_core_event_resolv_add_query (char *address);

/* for all events coming from the UDP listener */
PPL_DECLARE (int)
psp_core_event_add_sip_message (sipevent_t * evt);
PPL_DECLARE (int)
psp_core_event_add_imp_ack (sip_t * ack);

PPL_DECLARE (int)
psp_core_event_add_imp_traffic (transaction_t * inc_tr);

PPL_DECLARE (int)
psp_core_event_add_uas_request (psp_req_t * req);

PPL_DECLARE (int)
psp_core_event_add_sfull_request (psp_req_t * req);
PPL_DECLARE (int)
psp_core_event_add_sfull_cancel (transaction_t * tr_cancel);

#endif
