/*
  The partysip program is a modular SIP proxy server (SIP -rfc3261-)
  Copyright (C) 2002  WellX Telecom   - <partysip@wellx.com>
  Copyright (C) 2002  Aymeric MOIZARD - <jack@atosc.org>
  
  Partysip is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  Partysip is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with Foobar; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <partysip/partysip.h>
#include <partysip/psp_req.h>
#include <partysip/osip_msg.h>

#include <ppl/ppl_md5.h>

psp_core_t *core = NULL;

psp_core_t *
psp_core_get ()
{
  return core;
}

int psp_core_parse_servername (char *buf);
int psp_core_parse_serverip (char *buf);

int
psp_core_init ()
{
  char *tmp;
  int i;

  core = (psp_core_t *) smalloc (sizeof (psp_core_t));
  if (core == NULL)
    return -1;

  /* a malloc() (smalloc() is basically a malloc() wrapper
   * does not necessarily clear the malloc'd memory. some
   * implementations do but some don't.
   */
  memset (core, 0, sizeof (psp_core_t));

#ifdef DEFAULT_T1
  i = psp_osip_init (&(core->psp_osip), DEFAULT_T1);
#else
  i = psp_osip_init (&(core->psp_osip), 500);
#endif
  if (i != 0)
    goto mci_error1;

  i = pspm_resolv_init (&(core->resolv));
  if (i != 0)
    goto mci_error2;

  i = pspm_tlp_init (&(core->tlp));
  if (i != 0)
    goto mci_error3;

  i = pspm_imp_init (&(core->imp));
  if (i != 0)
    goto mci_error4;

  i = pspm_uap_init (&(core->uap));
  if (i != 0)
    goto mci_error5;

  i = pspm_sfp_init (&(core->sfp));
  if (i != 0)
    goto mci_error6;

  tmp = psp_config_get_element ("serverip");
  if (tmp == NULL)
    goto mci_error7;
  if (psp_core_parse_serverip (tmp) != 0)
    goto mci_error7;

  tmp = psp_config_get_element ("servername");
  if (tmp == NULL)
    goto mci_error8;
  if (psp_core_parse_servername (tmp) != 0)
    goto mci_error8;

  core->nat_enabled = 0;	/* only allowed on linux */
  core->ext_ip = NULL;
  core->ext_fqdn = NULL;
  core->ext_mask = NULL;
  core->lan_ip = NULL;
  core->lan_fqdn = NULL;
  core->lan_mask = NULL;
  for (i=0;i<PSP_NAT_MAX_CALL;i++)
    {
      *(core->fw_entries[i].hash) = '\0';
      *(core->fw_entries[i].remove_rule) = '\0';
      core->fw_entries[i].birth_date = 0;
      core->fw_entries[i].call_length = 0;
    }

#if defined(ALPHA_FIREWALL)
  tmp = psp_config_get_element ("nat");
  if (tmp != NULL && 0 == strcmp (tmp, "on"))
    {
      char *tmp_name;
      char *tmp_ip;
      char *tmp_mask;
      core->nat_enabled = 1;	/* enable NAT feature */
      tmp = psp_config_get_element ("if_ext");
      if (tmp == NULL)
	goto mci_error9;
      if (0 != ppl_dns_get_local_fqdn (&tmp_name, &tmp_ip, &tmp_mask, tmp))
	goto mci_error9;
      core->ext_ip = tmp_ip;
      core->ext_fqdn = tmp_name;
      core->ext_mask = tmp_mask;
      if (tmp_mask == NULL)
	goto mci_error10;
      tmp = psp_config_get_element ("if_lan");
      if (tmp == NULL)
	goto mci_error10;
      if (0 != ppl_dns_get_local_fqdn (&tmp_name, &tmp_ip, &tmp_mask, tmp))
	goto mci_error10;
      core->lan_ip = tmp_ip;
      core->lan_fqdn = tmp_name;
      core->lan_mask = tmp_mask;
      if (tmp_mask == NULL)
	goto mci_error11;
      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO2, NULL,
			      "NAT enabled %s(public)/%s -> %s(internal)/%s\n",
			      core->ext_ip, core->ext_mask,
			      core->lan_ip, core->lan_mask));
    }
#else
  /* no support for NAT */
  tmp = psp_config_get_element ("nat");
  if (tmp != NULL && 0 == strcmp (tmp, "on"))
    {
      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO2, NULL,
			      "There is no support for NAT (except on linux)\n"));
    }
#endif

  tmp = psp_config_get_element ("serverport_udp");
  if (tmp == NULL)
    core->port = 5060;
  else
    {
      char *end;
      end = strchr (tmp, ' ');
      if (end != NULL)
	*end = '\0';		/* discard any additionnal element */
      core->port = satoi (tmp);
    }

  tmp = psp_config_get_element ("disable_check_for_to_tag_in_cancel");
  if (tmp != NULL && 0 == strcmp (tmp, "on"))
    core->disable_check_for_to_tag_in_cancel = 1;

  tmp = psp_config_get_element ("rfc3327");
  if (tmp != NULL && 0 == strcmp (tmp, "on"))
    core->rfc3327 = 1;

  core->gl_lock = smutex_init ();
  if (core->gl_lock == NULL)
    goto mci_error11;

  return 0;

mci_error11:
#if defined(ALPHA_FIREWALL)
  sfree (core->lan_ip);
  sfree (core->lan_fqdn);
  sfree (core->lan_mask);
mci_error10:
  sfree (core->ext_ip);
  sfree (core->ext_fqdn);
  sfree (core->ext_mask);
mci_error9:
#endif
  for (i = 0; i < MAX_SERVER_IDENTIFIERS && core->servername[i] != NULL; i++)
    {
      sfree (core->servername[i]);
      core->servername[i] = NULL;
    }
mci_error8:
  for (i = 0; i < MAX_SERVER_IDENTIFIERS && core->serverip[i] != NULL; i++)
    {
      sfree (core->serverip[i]);
      core->serverip[i] = NULL;
    }
mci_error7:
  pspm_sfp_free (core->sfp);
  sfree (core->sfp);
mci_error6:
  pspm_uap_free (core->uap);
  sfree (core->uap);
mci_error5:
  pspm_imp_free (core->imp);
  sfree (core->imp);
mci_error4:
  pspm_tlp_free (core->tlp);
  sfree (core->tlp);
mci_error3:
  pspm_resolv_free (core->resolv);
  sfree (core->resolv);
mci_error2:
  psp_osip_free (core->psp_osip);
  sfree (core->psp_osip);
mci_error1:
  sfree (core);
  core = NULL;
  return -1;
}

int
psp_core_parse_servername (char *buf)
{
  char *c;
  size_t token_len = 0;
  int n = 0;

  c = buf;
  while (*c != '\0')
    {
      token_len = strcspn (c, " ,");
      if (token_len == 0)
	{
	  /* the token is only separators
	   * (example, buf = 'foo.example.com, bar.example.com'
	   *  and we are now here : ---------^
	   */
	  c++;
	}
      else
	{
	  if (n <= MAX_SERVER_IDENTIFIERS)
	    {
	      int need_bytes = token_len + 1;

	      /* now we need to check if there are any trailing dots
	       * that we should strip.
	       */
	      while (*(c + need_bytes - 2) == '.')
		need_bytes--;

	      /* XXX check if this token resolves? */

	      core->servername[n] = smalloc (need_bytes);
	      if (core->servername[n] == NULL)
		return -1;

	      sstrncpy ((char *) core->servername[n], c, need_bytes - 1);
	      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO2, NULL,
				      "servername token %i: '%s'\n", n,
				      core->servername[n]));
	      c += token_len;
	    }
	  else
	    {
	      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_ERROR, NULL,
				      "more than %i servernames\n",
				      MAX_SERVER_IDENTIFIERS));
	      return -1;
	    }
	  n++;
	}
    }
  return 0;			/* OK */
}

int
psp_core_parse_serverip (char *buf)
{
  char *c;
  size_t token_len = 0;
  int n = 0;

  c = buf;
  while (*c != '\0')
    {
      token_len = strcspn (c, " ,");
      if (token_len == 0)
	{
	  /* the token is only separators, just advance past it */
	  c++;
	}
      else
	{
	  if (n <= MAX_SERVER_IDENTIFIERS)
	    {
	      int need_bytes = token_len + 1;

	      core->serverip[n] = smalloc (need_bytes);
	      if (core->serverip[n] == NULL)
		return -1;

	      sstrncpy ((char *) core->serverip[n], c, need_bytes - 1);
	      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO2, NULL,
				      "serverip token %i: '%s'\n", n,
				      core->serverip[n]));
	      c += token_len;

	      if (inet_addr (core->serverip[n]) == -1)
		{
		  OSIP_TRACE (osip_trace
			      (__FILE__, __LINE__, OSIP_ERROR, NULL,
			       "serverip '%s' is not a valid IP address\n",
			       core->serverip[n]));
		  return -1;
		}
	    }
	  else
	    {
	      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_ERROR, NULL,
				      "more than %i serverips\n",
				      MAX_SERVER_IDENTIFIERS));
	      return -1;
	    }
	  n++;
	}
    }
  return 0;			/* OK */
}

void *
mythread_resolv (pspm_resolv_t * resolv)
{
  int i;

  i = pspm_resolv_execute (resolv, 5, 0, -1);	/* infinite loop */
  if (i != 0)
    return NULL;
  return NULL;
}

void *
mythread_tlp (pspm_tlp_t * tlp)
{
  int i;

  i = pspm_tlp_execute (tlp, 5, 0, -1);	/* infinite loop */
  if (i != 0)
    return NULL;
  return NULL;
}

void *
mythread_imp (pspm_imp_t * imp)
{
  int i;

  i = pspm_imp_execute (imp, 5, 0, -1);	/* infinite loop */
  if (i != 0)
    return NULL;
  return NULL;
}

void *
mythread_uap (pspm_uap_t * uap)
{
  int i;

  i = pspm_uap_execute (uap, 5, 0, -1);	/* infinite loop */
  if (i != 0)
    return NULL;
  return NULL;
}

void *
mythread_sfp (pspm_sfp_t * sfp)
{
  int i;

  i = pspm_sfp_execute (sfp, 5, 0, -1);	/* infinite loop */
  if (i != 0)
    return NULL;
  return NULL;
}

int
psp_core_start (int use_this_thread)
{
  int i;

  i = pspm_resolv_start (core->resolv, (void *(*)(void *)) mythread_resolv,
			 core->resolv);
  if (i != 0)
    return -1;

  i = pspm_tlp_start (core->tlp, (void *(*)(void *)) mythread_tlp, core->tlp);
  if (i != 0)
    return -1;

  i = pspm_imp_start (core->imp, (void *(*)(void *)) mythread_imp, core->imp);
  if (i != 0)
    goto pcs_error1;

  i = pspm_uap_start (core->uap, (void *(*)(void *)) mythread_uap, core->uap);
  if (i != 0)
    goto pcs_error2;

  if (use_this_thread == 1)	/* interactive mode: start a new thread as usual */
    {
      i =
	pspm_sfp_start (core->sfp, (void *(*)(void *)) mythread_sfp,
			core->sfp);
      if (i != 0)
	goto pcs_error3;
    }
  else
    {
      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO2, NULL,
			      "starting the sfp module with the main thread\n"));
      mythread_sfp (core->sfp);
      /* will never return until partysip is closed */
      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO1, NULL,
			      "shuting down partysip\n"));
    }

  return 0;
pcs_error3:
pcs_error2:
pcs_error1:

  return -1;
}

int
psp_core_execute ()
{
  int i;

  i = pspm_resolv_execute (core->resolv, 0, 0, 1);
  if (i != 0)
    goto pce_error1;

  i = pspm_tlp_execute (core->tlp, 0, 0, 1);
  if (i != 0)
    goto pce_error1;

  i = pspm_imp_execute (core->imp, 0, 0, 1);
  if (i != 0)
    goto pce_error1;

  i = pspm_uap_execute (core->uap, 0, 0, 1);
  if (i != 0)
    goto pce_error2;

  i = pspm_sfp_execute (core->sfp, 0, 0, 1);
  if (i != 0)
    goto pce_error3;

  return 0;
pce_error3:
pce_error2:
pce_error1:

  return -1;
}

void
psp_core_free ()
{
  int i;

  if (core == NULL)
    return;
  /* first: shutdown threads */
  psp_osip_release (core->psp_osip);	/* all transaction threads will die */
  pspm_resolv_release (core->resolv);	/* The resolver thread will die */
  pspm_tlp_release (core->tlp);	/* The tlp thread will die */
  pspm_imp_release (core->imp);	/* The imp thread will die */
  pspm_uap_release (core->uap);	/* The uap thread will die */

  if (core->sfp == NULL || core->sfp->module == NULL)
    return;
  if (core->sfp->module->thread == NULL)	/* this is not a thread! */
    {
      char q[2] = "q";
      i = ppl_pipe_write (core->sfp->module->wakeup, &q, 1);
      if (i != 1)
	{
	  OSIP_TRACE (osip_trace
		      (__FILE__, __LINE__, OSIP_ERROR, NULL,
		       "could not write in pipe!\n"));
	  return;
	}
      susleep (3000000);	/* as it's not a thread, we don't know
				   when this module will exit */
    }
  else
    {
      pspm_sfp_release (core->sfp);	/* The sfp thread will die */
    }

  /* done */

  /* here NO thread should be active */

  /* unregister other plugins and their methods */
  psp_core_remove_all_hook ();
  /* remove all plugins */
  psp_core_remove_all_plugins ();

  pspm_resolv_free (core->resolv);
  sfree (core->resolv);

  pspm_tlp_free (core->tlp);
  sfree (core->tlp);

  pspm_imp_free (core->imp);
  sfree (core->imp);
  pspm_uap_free (core->uap);
  sfree (core->uap);
  pspm_sfp_free (core->sfp);
  sfree (core->sfp);

  psp_osip_free (core->psp_osip);
  sfree (core->psp_osip);

  for (i = 0; i < MAX_SERVER_IDENTIFIERS && core->servername[i] != NULL; i++)
    {
      sfree (core->servername[i]);
      core->servername[i] = NULL;
    }
  for (i = 0; i < MAX_SERVER_IDENTIFIERS && core->serverip[i] != NULL; i++)
    {
      sfree (core->serverip[i]);
      core->serverip[i] = NULL;
    }

  smutex_destroy (core->gl_lock);
  sfree (core->gl_lock);

  sfree (core);
}

int
psp_core_lock ()
{
  return smutex_lock (core->gl_lock);
}

int
psp_core_unlock ()
{
  return smutex_unlock (core->gl_lock);
}

void
psp_core_remove_all_plugins ()
{
  {
    tlp_plugin_t *p;

    for (p = core->tlp->tlp_plugins; p != NULL; p = core->tlp->tlp_plugins)
      {
	core->tlp->tlp_plugins = p->next;
	tlp_plugin_free (p);
	sfree (p);
      }
  }
  {
    imp_plugin_t *p;

    for (p = core->imp->imp_plugins; p != NULL; p = core->imp->imp_plugins)
      {
	REMOVE_ELEMENT (core->imp->imp_plugins, p);
	imp_plugin_free (p);
	sfree (p);
      }
  }
  {
    uap_plugin_t *p;

    for (p = core->uap->uap_plugins; p != NULL; p = core->uap->uap_plugins)
      {
	REMOVE_ELEMENT (core->uap->uap_plugins, p) uap_plugin_free (p);
	sfree (p);
      }
  }
  {
    sfp_plugin_t *p;

    for (p = core->sfp->sfp_plugins; p != NULL; p = core->sfp->sfp_plugins)
      {
	REMOVE_ELEMENT (core->sfp->sfp_plugins, p);
	sfp_plugin_free (p);
	sfree (p);
      }
  }
}

void
psp_core_remove_all_hook ()
{
  /* remove imp hooks */
  imp_func_tab_free (core->imp->func_invites);
  imp_func_tab_free (core->imp->func_registers);
  imp_func_tab_free (core->imp->func_byes);
  imp_func_tab_free (core->imp->func_optionss);
  imp_func_tab_free (core->imp->func_infos);
  imp_func_tab_free (core->imp->func_cancels);
  imp_func_tab_free (core->imp->func_notifys);
  imp_func_tab_free (core->imp->func_subscribes);
  imp_func_tab_free (core->imp->func_unknowns);

  uap_func_tab_free (core->uap->func_1xxs);
  uap_func_tab_free (core->uap->func_2xxs);
  uap_func_tab_free (core->uap->func_3xxs);
  uap_func_tab_free (core->uap->func_4xxs);
  uap_func_tab_free (core->uap->func_5xxs);
  uap_func_tab_free (core->uap->func_6xxs);

  sfp_fwd_func_tab_free (core->sfp->fwd_invites);
  sfp_fwd_func_tab_free (core->sfp->fwd_registers);
  sfp_fwd_func_tab_free (core->sfp->fwd_byes);
  sfp_fwd_func_tab_free (core->sfp->fwd_optionss);
  sfp_fwd_func_tab_free (core->sfp->fwd_infos);
  sfp_fwd_func_tab_free (core->sfp->fwd_cancels);
  sfp_fwd_func_tab_free (core->sfp->fwd_notifys);
  sfp_fwd_func_tab_free (core->sfp->fwd_subscribes);
  sfp_fwd_func_tab_free (core->sfp->fwd_unknowns);

  sfp_rcv_func_tab_free (core->sfp->rcv_1xxs);
  sfp_rcv_func_tab_free (core->sfp->rcv_2xxs);
  sfp_rcv_func_tab_free (core->sfp->rcv_3xxs);
  sfp_rcv_func_tab_free (core->sfp->rcv_4xxs);
  sfp_rcv_func_tab_free (core->sfp->rcv_5xxs);
  sfp_rcv_func_tab_free (core->sfp->rcv_6xxs);

  sfp_snd_func_tab_free (core->sfp->snd_1xxs);
  sfp_snd_func_tab_free (core->sfp->snd_2xxs);
  sfp_snd_func_tab_free (core->sfp->snd_3xxs);
  sfp_snd_func_tab_free (core->sfp->snd_4xxs);
  sfp_snd_func_tab_free (core->sfp->snd_5xxs);
  sfp_snd_func_tab_free (core->sfp->snd_6xxs);

}

char *
psp_core_sfp_generate_branch_for_request (sip_t * request)
{
  char *branch;

  branch = psp_core_default_generate_branch_for_request (request);
  if (branch == NULL)
    return NULL;
  branch[7] = 'f';
  return branch;
}

char *
psp_core_default_generate_branch_for_request (sip_t * request)
{
  via_t *via;
  generic_param_t *b;
  char *branch;

  branch = (char *) smalloc (44);	/* Magic cookie + 'f' or 'l'+hash+'.xx'+'\0' */

  /* check if the request has the magic cookie for compliance with latest rfc */
  via = list_get (request->vias, 0);
  via_param_getbyname (via, "branch", &b);
  if (b != NULL && 0 == strncmp ("z9hG4bK", b->gvalue, 7))
    {				/* compute hash with branchid */
      MD5_CTX Md5Ctx;
      HASH HA1;

      ppl_MD5Init (&Md5Ctx);
      ppl_MD5Update (&Md5Ctx, (unsigned char *) b->gvalue,
		     strlen (b->gvalue));
      ppl_MD5Final ((unsigned char *) HA1, &Md5Ctx);

      sstrncpy (branch, "z9hG4bKl", 8);
      ppl_md5_hash_to_hex (HA1, branch + 8);
    }
  else
    {				/* not compliant...: hash is more complex */
      char *topmostvia;
      char *cid;
      char *rquri;
      int i;
      generic_param_t *tag;
      MD5_CTX Md5Ctx;
      HASH HA1;

      ppl_MD5Init (&Md5Ctx);

      i = via_2char (via, &topmostvia);
      ppl_MD5Update (&Md5Ctx, (unsigned char *) topmostvia,
		     strlen (topmostvia));
      sfree (topmostvia);
      ppl_MD5Update (&Md5Ctx, (unsigned char *) ":", 1);

      to_get_tag (request->to, &tag);
      if (tag != NULL)
	{
	  ppl_MD5Update (&Md5Ctx, (unsigned char *) tag->gvalue,
			 strlen (tag->gvalue));
	}
      ppl_MD5Update (&Md5Ctx, (unsigned char *) ":", 1);

      from_get_tag (request->from, &tag);
      if (tag != NULL)
	{
	  ppl_MD5Update (&Md5Ctx, (unsigned char *) tag->gvalue,
			 strlen (tag->gvalue));
	}
      else
	{
	  OSIP_TRACE (osip_trace
		      (__FILE__, __LINE__, OSIP_WARNING, NULL,
		       "remote UA is not compliant: no tag in the from\n"));
	}
      ppl_MD5Update (&Md5Ctx, (unsigned char *) ":", 1);

      call_id_2char (request->call_id, &cid);
      ppl_MD5Update (&Md5Ctx, (unsigned char *) cid, strlen (cid));
      sfree (cid);
      ppl_MD5Update (&Md5Ctx, (unsigned char *) ":", 1);

      ppl_MD5Update (&Md5Ctx, (unsigned char *) request->cseq->number,
		     strlen (request->cseq->number));
      ppl_MD5Update (&Md5Ctx, (unsigned char *) ":", 1);

      url_2char (request->strtline->rquri, &rquri);
      ppl_MD5Update (&Md5Ctx, (unsigned char *) rquri, strlen (rquri));
      sfree (rquri);
      ppl_MD5Final ((unsigned char *) HA1, &Md5Ctx);

      sstrncpy (branch, "z9hG4bKl", 8);
      ppl_md5_hash_to_hex (HA1, branch + 8);
      OSIP_TRACE (osip_trace
		  (__FILE__, __LINE__, OSIP_INFO4, NULL,
		   "Branch hash (remote client is not compliant): %s\n",
		   branch));
    }
  return branch;
}

int
psp_core_sfp_new_rsp_ctx (rsp_ctx_t ** ctx, psp_req_t * req)
{
  rsp_ctx_t *rsp_ctx;
  char *branch;
  int i;

  *ctx = NULL;
  branch = psp_core_sfp_generate_branch_for_request (req->request);
  if (branch == NULL)
    return -1;
  i = rsp_ctx_init (&rsp_ctx, branch);
  if (i != 0)
    {
      sfree (branch);
      return -1;
    }
  *ctx = rsp_ctx;
  return 0;
}

transaction_t *
psp_core_uap_create_transaction (sipevent_t * evt)
{
  return osip_create_transaction (core->psp_osip->osip, evt);
}

PPL_DECLARE (char *) psp_core_get_ip_of_local_proxy ()
{
  unsigned long int one_inet_addr;

  if ((int) (one_inet_addr = inet_addr (core->serverip[0])) == -1)
    {
      if ((int) (one_inet_addr = inet_addr (core->servername[0])) == -1)
	return core->serverip[0];
      return core->servername[0];
    }
  return core->serverip[0];
}

PPL_DECLARE (int) psp_core_get_port_of_local_proxy ()
{
  return core->port;
}

PPL_DECLARE (int) psp_core_is_responsible_for_this_domain (url_t * url)
{
  /*
     int is_an_ip, i;
     int match = 0;

     if (url == NULL)
     return -1;

     if (0!=strcmp(url->scheme, "sip")
     &&0!=strcmp(url->scheme, "sips"))
     return -1;

     is_an_ip = inet_addr (url->host);
     if (is_an_ip == -1)
     {
     for (i = 0; i < MAX_SERVER_IDENTIFIERS && core->servername[i] != NULL; i++)
     {
     #ifndef WIN32
     if (!strcasecmp (url->host, core->servername[i]))
     {
     match = 1;
     OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO2, NULL,
     "psp_core_is_responsible_for_this_route() '%s' == '%s'\n",
     url->host, core->servername[i]));
     return 0;
     }
     #else
     if (!stricmp (url->host, core->servername[i]))
     {
     match = 1;
     OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO2, NULL,
     "psp_core_is_responsible_for_this_route() '%s' == '%s'\n",
     url->host, core->servername[i]));
     return 0;
     }
     #endif
     }
     } else
     {
     for (i = 0; i < MAX_SERVER_IDENTIFIERS && core->serverip[i] != NULL; i++)
     {
     if (!strcmp (url->host, core->serverip[i]))
     {
     match = 1;
     OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO2, NULL,
     "psp_core_is_responsible_for_this_route() '%s' == '%s'\n",
     url->host, core->serverip[i]));
     return 0;
     }
     }
     }

     return -1;
   */
  return psp_core_is_responsible_for_this_request_uri (url);
}

PPL_DECLARE (int) psp_core_is_responsible_for_this_request_uri (url_t * url)
{
  int port, is_an_ip, i;
  int match = 0;

  if (url == NULL)
    return -1;

  if (0 != strcmp (url->scheme, "sip") && 0 != strcmp (url->scheme, "sips"))
    return -1;			/* only sip url are allowed here */

  is_an_ip = inet_addr (url->host);
  if (is_an_ip == -1)
    {
      /* is a hostname, search our list of names for this machine */

      for (i = 0; i < MAX_SERVER_IDENTIFIERS && core->servername[i] != NULL;
	   i++)
	{
#ifndef WIN32
	  if (!strcasecmp (url->host, core->servername[i]))
	    {
	      match = 1;
	      break;
	    }
#else
	  if (!stricmp (url->host, core->servername[i]))
	    {
	      match = 1;
	      break;
	    }
#endif
	}
    }
  else
    {
      /* is an IP address, search our list of IP addresses for this machine */

      for (i = 0; i < MAX_SERVER_IDENTIFIERS && core->serverip[i] != NULL;
	   i++)
	{
	  if (!strcmp (url->host, core->serverip[i]))
	    {
	      match = 1;
	      break;
	    }
	}
    }


  if (match)
    {
      /* If more than one proxy runs on this host, we have to check
         for the port. */

      /*  if the rquri has not port specified
         then this is a request for this domain */
      if (url->port == NULL)
	{
	  OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO4, NULL,
				  "psp_core_is_responsible_for_this_domain() '%s'\n",
				  url->host));
	  return 0;
	}

      if (core->port == 5060)
	{
	  /* in this case, the port MUST be the same than the one
	     used by the proxy */
	  if (url->port == NULL)
	    {
	      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO4, NULL,
				      "psp_core_is_responsible_for_this_domain() '%s'\n",
				      url->host));
	      return 0;
	    }
	  port = satoi (url->port);
	  if (5060 == port)
	    {
	      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO4, NULL,
				      "psp_core_is_responsible_for_this_domain() '%s'\n",
				      url->host));
	      return 0;
	    }
	  return -1;
	}
      else
	{
	  /* if the port is specified in the domain, it can
	     be the same. */
	  port = satoi (url->port);
	  if (core->port == port)
	    {
	      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO4, NULL,
				      "psp_core_is_responsible_for_this_domain() '%s'\n",
				      url->host));
	      return 0;
	    }
	  return -1;
	}
      return -1;
    }

  return -1;
  /*  return psp_core_is_responsible_for_this_route (url); */
}

PPL_DECLARE (int) psp_core_is_responsible_for_this_route (url_t * url)
{
  int port, is_an_ip, i;
  int match = 0;

  if (url == NULL)
    return -1;

  if (0 != strcmp (url->scheme, "sip") && 0 != strcmp (url->scheme, "sips"))
    return -1;			/* only sip url are allowed here */

  is_an_ip = inet_addr (url->host);
  if (is_an_ip == -1)
    {
      /* is a hostname, search our list of names for this machine */

      for (i = 0; i < MAX_SERVER_IDENTIFIERS && core->servername[i] != NULL;
	   i++)
	{
#ifndef WIN32
	  if (!strcasecmp (url->host, core->servername[i]))
	    {
	      match = 1;
	      break;
	    }
#else
	  if (!stricmp (url->host, core->servername[i]))
	    {
	      match = 1;
	      break;
	    }
#endif
	}
    }
  else
    {
      /* is an IP address, search our list of IP addresses for this machine */

      for (i = 0; i < MAX_SERVER_IDENTIFIERS && core->serverip[i] != NULL;
	   i++)
	{
	  if (!strcmp (url->host, core->serverip[i]))
	    {
	      match = 1;
	      break;
	    }
	}
    }

  if (match)
    {
      /* If more than one proxy runs on this host, we have to check
         for the port. */
      if (url->port != NULL)
	port = satoi (url->port);
      else
	port = 5060;
      if (core->port == port)
	{
	  OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO4, NULL,
				  "psp_core_is_responsible_for_this_route() '%s'\n",
				  url->host));
	  return 0;
	}
      return -1;
    }
  return -1;
}


PPL_DECLARE (int) psp_core_find_transaction_and_add_event (sipevent_t * evt)
{
  return osip_find_transaction_and_add_event (core->psp_osip->osip, evt);
}

/*
transaction_t *
psp_core_osip_find_transaction(sipevent_t *evt)
{
  return osip_find_transaction(core->psp_osip->osip, evt);
}
*/

int
psp_core_cb_snd_message (transaction_t * tr, sip_t * sip, char *host,
			 int port, int out_socket)
{
  if (MSG_IS_RESPONSE (sip) && host == NULL)
    {				/* automatic search of destination */
      /* if (reliable_transport) -> use existing connection
         .                     else use received
         .                     else use sent-by
         elseif (maddr is found) use the maddr(+use the indicated ttl if multicast)
         elseif (unreliable protocol) use received
         .                       else use sent-by        
       */
      via_t *via;
      generic_param_t *maddr;
      generic_param_t *received;
      generic_param_t *rport;
      via = list_get (sip->vias, 0);
      if (via == NULL)
	return -2;
      via_param_getbyname (via, "maddr", &maddr);
      via_param_getbyname (via, "received", &received);
      via_param_getbyname (via, "rport", &rport);
      /* 1: skip the TCP implementation, partysip don't support it by now */
      /* 2: check maddr and multicast usage */
      if (maddr != NULL)
	host = maddr->gvalue;
      /* we should check if this is a multicast address and use
         set the "ttl" in this case. (this must be done in the
         UDP message (not at the SIP layer) */
      else if (received != NULL)
	host = received->gvalue;
      else
	host = via->host;

      if (rport == NULL || rport->gvalue == NULL)
	{
	  if (via->port != NULL)
	    port = satoi (via->port);
	  else
	    port = 5060;
	}
      else
	port = satoi (rport->gvalue);
    }
  return pspm_tlp_send_message (core->tlp, tr, sip, host, port, out_socket);
}

int
psp_core_switch_to_uas_mode (psp_req_t * req, int status)
{
  req->uas_status = status;
  psp_core_event_add_uas_request (req);
  return 0;
}



/* Useless?
psp_req_t *
psp_core_sfp_find_req (int osip_id)
{
  psp_req_t *mr;
  sfp_branch_t *sfp;

  for (mr = core->sfp->psp_requests; mr != NULL; mr = mr->next)
    {
      if (mr->rsp_ctx != NULL)
        {
          for (sfp = mr->rsp_ctx->branches; sfp != NULL; sfp = sfp->next)
            {
              if (sfp->inc_tr!=NULL && sfp->inc_tr->transactionid == osip_id)
                {
                  OSIP_TRACE (osip_trace
                              (__FILE__, __LINE__, OSIP_INFO3, NULL,
                               "found branch %i for sfp module!\n",
			       mr->inc_tr->transactionid));
                  return mr;
                }
            }
        }
    }
  return NULL;
}
 */


  /* the strict router issue: TODO:
     "The proxy MUST inspect the Request-URI of the request.  If the
     Request-URI of the request contains a value this proxy previously
     placed into a Record-Route header field (see Section 16.6 item 4),
     the proxy MUST replace the Request-URI in the request with the last
     value from the Route header field, and remove that value from the
     Route header field. The proxy MUST then proceed as if it received
     this modified request.
     ...
     This requirement does not obligate a proxy to keep state in
     order to detect URIs it previously placed in Record-Route
     header fields. Instead, a proxy need only place enough
     information in those URIs to recognize them as values it
     provided when they later appear."
   */

PPL_DECLARE (int) psp_core_fix_strict_router_issue (sipevent_t * evt)
{
  url_param_t *psp_param;

  if (MSG_IS_RESPONSE (evt->sip))
    return 0;

  url_uparam_getbyname (evt->sip->strtline->rquri, "psp", &psp_param);
  if (psp_param != NULL && !list_eol (evt->sip->routes, 0))
    {				/* !! strict rooter detected! (compliant with old draft...)
				   We have to rewrite the request-uri and routes */
      route_t *route;
      char *magicstring;
      int pos;

      magicstring = psp_config_get_element ("magicstring");	/* Who I am? */
      if (0 == strcmp (magicstring, psp_param->gvalue))
	{			/* Yes, it's mine */
	  /* replace the rq-uri with the last route header */
	  int i;
	  url_t *oldurl;

	  pos = 0;
	  while (!list_eol (evt->sip->routes, pos))
	    pos++;
	  pos--;
	  msg_getroute (evt->sip, pos, &route);
	  list_remove (evt->sip->routes, pos);
	  oldurl = evt->sip->strtline->rquri;

	  evt->sip->strtline->rquri = route->url;
	  route->url = NULL;	/* avoid double free */
	  route_free (route);
	  sfree (route);

	  i = route_init (&route);
	  if (i != 0)
	    return -1;
	  route->url = oldurl;
	  list_add (evt->sip->routes, route, 0);
	}
    }
  /* request is now compliant to the latest draft :-) */

  /* now we should check if the request-uri has an maddr paramater...
     (I won't receive such request until I add maddr and partysip-0.5.0 don't!)
   */
  return 0;
}

int
psp_core_handle_late_answer (sip_t * answer)
{
  int i;
  /*  int port;
     via_t *via; */
  sip_t *answerxxx;

  i = osip_msg_sfp_build_response_to_forward (&answerxxx, answer);
  if (i != 0)
    {
      OSIP_TRACE (osip_trace
		  (__FILE__, __LINE__, OSIP_ERROR, NULL,
		   "imp module: Can't build the late answer to forward!\n"));
      return -1;
    }

  /*
     via = list_get (answerxxx->vias, 0);
     if (via == NULL)
     {
     msg_free (answerxxx);
     sfree (answerxxx);
     return -1;
     }
     if (via->port == NULL)
     port = 5060;
     else
     port = atoi (via->port);
     psp_core_cb_snd_message (NULL, answerxxx, via->host, port, -1);
   */
  psp_core_cb_snd_message (NULL, answerxxx, NULL, 5060, -1);

  msg_free (answerxxx);
  sfree (answerxxx);

  return 0;
}
