/*
  The partysip program is a modular SIP proxy server (SIP -rfc3261-)
  Copyright (C) 2002  WellX Telecom   - <partysip@wellx.com>
  Copyright (C) 2002  Aymeric MOIZARD - <jack@atosc.org>
  
  Partysip is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  Partysip is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with Foobar; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <partysip/partysip.h>

#include <partysip/osip_msg.h>

#include <ppl/ppl_dns.h>

extern psp_core_t *core;

int
pspm_sfp_init (pspm_sfp_t ** pspm)
{
  int i;

  (*pspm) = (pspm_sfp_t *) smalloc (sizeof (pspm_sfp_t));
  if (*pspm == NULL)
    return -1;

  i = module_init (&((*pspm)->module), MOD_THREAD, "sfp");
  if (i != 0)
    goto psi_error1;
  (*pspm)->sfp_plugins = NULL;


  /*
     (*pspm)->mut_psp_requests = smutex_init ();
     if ((*pspm)->mut_psp_requests == NULL)
     goto psi_error2;
   */
  (*pspm)->psp_requests = NULL;

  (*pspm)->sfull_request = (fifo_t *) smalloc (sizeof (fifo_t));
  if ((*pspm)->sfull_request == NULL)
    goto psi_error3;
  fifo_init ((*pspm)->sfull_request);

  (*pspm)->sfull_cancels = (fifo_t *) smalloc (sizeof (fifo_t));
  if ((*pspm)->sfull_cancels == NULL)
    goto psi_error4;
  fifo_init ((*pspm)->sfull_cancels);

  (*pspm)->broken_transactions = (list_t *) smalloc (sizeof (list_t));
  if ((*pspm)->broken_transactions == NULL)
    goto psi_error5;
  list_init ((*pspm)->broken_transactions);

  (*pspm)->psp_requests = NULL;

  (*pspm)->fwd_invites = NULL;
  (*pspm)->fwd_acks = NULL;
  (*pspm)->fwd_registers = NULL;
  (*pspm)->fwd_byes = NULL;
  (*pspm)->fwd_optionss = NULL;
  (*pspm)->fwd_infos = NULL;
  (*pspm)->fwd_cancels = NULL;
  (*pspm)->fwd_notifys = NULL;
  (*pspm)->fwd_subscribes = NULL;
  (*pspm)->fwd_unknowns = NULL;

  (*pspm)->rcv_1xxs = NULL;
  (*pspm)->rcv_2xxs = NULL;
  (*pspm)->rcv_3xxs = NULL;
  (*pspm)->rcv_4xxs = NULL;
  (*pspm)->rcv_5xxs = NULL;
  (*pspm)->rcv_6xxs = NULL;

  (*pspm)->snd_1xxs = NULL;
  (*pspm)->snd_2xxs = NULL;
  (*pspm)->snd_3xxs = NULL;
  (*pspm)->snd_4xxs = NULL;
  (*pspm)->snd_5xxs = NULL;
  (*pspm)->snd_6xxs = NULL;

  i = sfp_fwd_func_tab_init (&((*pspm)->fwd_invites));
  if (i != 0)
    goto psi_error6;
  i = sfp_fwd_func_tab_init (&((*pspm)->fwd_acks));
  if (i != 0)
    goto psi_error6;
  i = sfp_fwd_func_tab_init (&((*pspm)->fwd_registers));
  if (i != 0)
    goto psi_error6;
  i = sfp_fwd_func_tab_init (&((*pspm)->fwd_byes));
  if (i != 0)
    goto psi_error6;
  i = sfp_fwd_func_tab_init (&((*pspm)->fwd_optionss));
  if (i != 0)
    goto psi_error6;
  i = sfp_fwd_func_tab_init (&((*pspm)->fwd_infos));
  if (i != 0)
    goto psi_error6;
  i = sfp_fwd_func_tab_init (&((*pspm)->fwd_cancels));
  if (i != 0)
    goto psi_error6;
  i = sfp_fwd_func_tab_init (&((*pspm)->fwd_notifys));
  if (i != 0)
    goto psi_error6;
  i = sfp_fwd_func_tab_init (&((*pspm)->fwd_subscribes));
  if (i != 0)
    goto psi_error6;
  i = sfp_fwd_func_tab_init (&((*pspm)->fwd_unknowns));
  if (i != 0)
    goto psi_error6;

  i = sfp_rcv_func_tab_init (&((*pspm)->rcv_1xxs));
  if (i != 0)
    goto psi_error6;
  i = sfp_rcv_func_tab_init (&((*pspm)->rcv_2xxs));
  if (i != 0)
    goto psi_error6;
  i = sfp_rcv_func_tab_init (&((*pspm)->rcv_3xxs));
  if (i != 0)
    goto psi_error6;
  i = sfp_rcv_func_tab_init (&((*pspm)->rcv_4xxs));
  if (i != 0)
    goto psi_error6;
  i = sfp_rcv_func_tab_init (&((*pspm)->rcv_5xxs));
  if (i != 0)
    goto psi_error6;
  i = sfp_rcv_func_tab_init (&((*pspm)->rcv_6xxs));
  if (i != 0)
    goto psi_error6;

  i = sfp_snd_func_tab_init (&((*pspm)->snd_1xxs));
  if (i != 0)
    goto psi_error6;
  i = sfp_snd_func_tab_init (&((*pspm)->snd_2xxs));
  if (i != 0)
    goto psi_error6;
  i = sfp_snd_func_tab_init (&((*pspm)->snd_3xxs));
  if (i != 0)
    goto psi_error6;
  i = sfp_snd_func_tab_init (&((*pspm)->snd_4xxs));
  if (i != 0)
    goto psi_error6;
  i = sfp_snd_func_tab_init (&((*pspm)->snd_5xxs));
  if (i != 0)
    goto psi_error6;
  i = sfp_snd_func_tab_init (&((*pspm)->snd_6xxs));
  if (i != 0)
    goto psi_error6;

  return 0;

psi_error6:
  module_release ((*pspm)->module);
  pspm_sfp_free (*pspm);
  sfree (*pspm);
  return -1;

psi_error5:
  fifo_free ((*pspm)->sfull_cancels);
psi_error4:
  sfree ((*pspm)->sfull_cancels);
  fifo_free ((*pspm)->sfull_request);
psi_error3:
  /*
     sfree ((*pspm)->sfull_request);
     smutex_destroy ((*pspm)->mut_psp_requests);
     psi_error2: */

  module_release ((*pspm)->module);
  module_free ((*pspm)->module);
  sfree ((*pspm)->module);
psi_error1:
  sfree (*pspm);
  return -1;
}

void
pspm_sfp_free (pspm_sfp_t * pspm)
{
  sfp_plugin_t *p;
  psp_req_t *req;
  transaction_t *tr;

  if (pspm == NULL)
    return;

  sfp_fwd_func_tab_free (pspm->fwd_invites);
  sfp_fwd_func_tab_free (pspm->fwd_acks);
  sfp_fwd_func_tab_free (pspm->fwd_registers);
  sfp_fwd_func_tab_free (pspm->fwd_byes);
  sfp_fwd_func_tab_free (pspm->fwd_optionss);
  sfp_fwd_func_tab_free (pspm->fwd_infos);
  sfp_fwd_func_tab_free (pspm->fwd_cancels);
  sfp_fwd_func_tab_free (pspm->fwd_notifys);
  sfp_fwd_func_tab_free (pspm->fwd_subscribes);
  sfp_fwd_func_tab_free (pspm->fwd_unknowns);

  sfp_rcv_func_tab_free (pspm->rcv_1xxs);
  sfp_rcv_func_tab_free (pspm->rcv_2xxs);
  sfp_rcv_func_tab_free (pspm->rcv_3xxs);
  sfp_rcv_func_tab_free (pspm->rcv_4xxs);
  sfp_rcv_func_tab_free (pspm->rcv_5xxs);
  sfp_rcv_func_tab_free (pspm->rcv_6xxs);

  sfp_snd_func_tab_free (pspm->snd_1xxs);
  sfp_snd_func_tab_free (pspm->snd_2xxs);
  sfp_snd_func_tab_free (pspm->snd_3xxs);
  sfp_snd_func_tab_free (pspm->snd_4xxs);
  sfp_snd_func_tab_free (pspm->snd_5xxs);
  sfp_snd_func_tab_free (pspm->snd_6xxs);


  sfree (pspm->fwd_invites);
  sfree (pspm->fwd_acks);
  sfree (pspm->fwd_registers);
  sfree (pspm->fwd_byes);
  sfree (pspm->fwd_optionss);
  sfree (pspm->fwd_infos);
  sfree (pspm->fwd_cancels);
  sfree (pspm->fwd_notifys);
  sfree (pspm->fwd_subscribes);
  sfree (pspm->fwd_unknowns);

  sfree (pspm->rcv_1xxs);
  sfree (pspm->rcv_2xxs);
  sfree (pspm->rcv_3xxs);
  sfree (pspm->rcv_4xxs);
  sfree (pspm->rcv_5xxs);
  sfree (pspm->rcv_6xxs);

  sfree (pspm->snd_1xxs);
  sfree (pspm->snd_2xxs);
  sfree (pspm->snd_3xxs);
  sfree (pspm->snd_4xxs);
  sfree (pspm->snd_5xxs);
  sfree (pspm->snd_6xxs);

  module_free (pspm->module);
  sfree (pspm->module);

  req = fifo_tryget (pspm->sfull_request);
  while (req != NULL)
    {
      psp_req_free (req);
      sfree (req);
      req = fifo_tryget (pspm->sfull_request);
    }

  fifo_free (pspm->sfull_request);
  sfree (pspm->sfull_request);

  tr = fifo_tryget (pspm->sfull_cancels);
  while (tr != NULL)
    {
      if (tr->state == NIST_TERMINATED)
	{
	  transaction_free2 (tr);
	  sfree (tr);
	}
      else
	{
	  transaction_free (tr);
	  sfree (tr);
	}
      tr = fifo_tryget (pspm->sfull_cancels);
    }

  fifo_free (pspm->sfull_cancels);
  sfree (pspm->sfull_cancels);

  tr = list_get (pspm->broken_transactions, 0);
  while (tr != NULL)
    {
      list_remove (pspm->broken_transactions, 0);
      if (tr->state == NICT_TERMINATED || tr->state == ICT_TERMINATED
	  || tr->state == NIST_TERMINATED || tr->state == IST_TERMINATED)
	{
	  transaction_free2 (tr);
	  sfree (tr);
	}
      else
	{
	  transaction_free (tr);
	  sfree (tr);
	}
      tr = list_get (pspm->broken_transactions, 0);
    }

  sfree (pspm->broken_transactions);

  /*
     smutex_destroy (pspm->mut_psp_requests);
     sfree (pspm->mut_psp_requests);
   */

  for (req = pspm->psp_requests; req != NULL; req = pspm->psp_requests)
    {
      REMOVE_ELEMENT (pspm->psp_requests, req);
      psp_req_free (req);
      sfree (req);
    }

  for (p = pspm->sfp_plugins; p != NULL; p = pspm->sfp_plugins)
    {
      REMOVE_ELEMENT (pspm->sfp_plugins, p);
      sfp_plugin_free (p);
      sfree (p);
    }

}


int
pspm_sfp_release (pspm_sfp_t * pspm)
{
  int i;

  i = module_release (pspm->module);
  if (i != 0)
    return -1;
  return 0;
}

int
pspm_sfp_start (pspm_sfp_t * pspm, void *(*func_start) (void *), void *arg)
{
  int i;

  i = module_start (pspm->module, func_start, arg);
  if (i != 0)
    return -1;
  return 0;
}

/*
int
pspm_sfp_lock_psp_requests_access (pspm_sfp_t * pspm)
{
  return smutex_lock (pspm->mut_psp_requests);
}

int
pspm_sfp_unlock_psp_requests_access (pspm_sfp_t * pspm)
{
  return smutex_unlock (pspm->mut_psp_requests);
}
*/

int
pspm_sfp_execute (pspm_sfp_t * pspm, int sec_max, int usec_max, int max)
{
  psp_req_t *req;
  psp_req_t *pnext;
  transaction_t *tr;

  fd_set sfp_fdset;
  int max_fd;
  struct timeval tv;

  if (pspm == NULL)
    return -1;

  while (1)
    {
      int s;
      int i;
      int locked;

      tv.tv_sec = sec_max;
      tv.tv_usec = usec_max;

      max--;
      FD_ZERO (&sfp_fdset);
      s = ppl_pipe_get_read_descr (pspm->module->wakeup);
      if (s <= 1)
	{
	  OSIP_TRACE (osip_trace
		      (__FILE__, __LINE__, OSIP_ERROR, NULL,
		       "No wakeup value in module sfp!\n"));
	  return -1;
	}
      FD_SET (s, &sfp_fdset);

      max_fd = s;

      if ((sec_max == -1) || (usec_max == -1))
	i = select (max_fd + 1, &sfp_fdset, NULL, NULL, NULL);
      else
	i = select (max_fd + 1, &sfp_fdset, NULL, NULL, &tv);

      if (FD_ISSET (s, &sfp_fdset))
	{
	  char tmp[51];

	  i = ppl_pipe_read (pspm->module->wakeup, tmp, 50);
	  if (i >= 1 && tmp[0] == 'q')
	    {
	      OSIP_TRACE (osip_trace
			  (__FILE__, __LINE__, OSIP_INFO2, NULL,
			   "sfp module: Exiting!\n"));
	      return 0;
	    }
	  else
	    {
	      OSIP_TRACE (osip_trace
			  (__FILE__, __LINE__, OSIP_INFO2, NULL,
			   "sfp module: wake up!\n"));
	    }
	}
      /* pspm_sfp_lock_psp_requests_access (pspm); */


      /***********************************************************/
      /* Prepare statefull context for new psp_req_t elements.   */
      req = fifo_tryget (pspm->sfull_request);
      while (req != NULL)
	{
	  /* psp_req element containing a request coming from imp */
	  pspm_sfp_handle_new_request (pspm, req);
	  req = fifo_tryget (pspm->sfull_request);
	}

      for (pnext = pspm->psp_requests, req = pnext; pnext != NULL;
	   req = pnext)
	{
	  pnext = req->next;
	  if (rsp_ctx_get_state (req->rsp_ctx) == RSPCTX_CALLING
	      || rsp_ctx_get_state (req->rsp_ctx) == RSPCTX_PROCEEDING)
	    {
	      pspm_sfp_handle_new_location (pspm, req);
	    }
	}

      locked = 0;
      /***********************************************************/
      /*    handle new incoming CANCEL requests                  */
      tr = fifo_tryget (pspm->sfull_cancels);
      if (tr != NULL)
	{
	  psp_core_lock ();
	  while (tr != NULL)
	    {
	      /* psp_req element containing a request coming from imp */
	      pspm_sfp_handle_new_cancel (pspm, tr);
	      tr = fifo_tryget (pspm->sfull_cancels);
	    }
	  locked = 1;
	}


      /***********************************************************/
      /*    handle events in RSPCTX_CALLING state                */
      /*               or in  RSPCTX_PROCEEDING state            */

      for (pnext = pspm->psp_requests, req = pnext; pnext != NULL;
	   req = pnext)
	{
	  pnext = req->next;
	  if (rsp_ctx_get_state (req->rsp_ctx) == RSPCTX_CALLING)
	    {
	      if (locked == 0)
		{
		  psp_core_lock ();
		  locked = 1;
		}
	      pspm_sfp_delete_broken_location (pspm, req);
	    }
	}

      /***********************************************************/
      /*    handle events in RSPCTX_PROCEEDING state             */
      for (pnext = pspm->psp_requests, req = pnext; pnext != NULL;
	   req = pnext)
	{
	  pnext = req->next;
	  if (rsp_ctx_get_state (req->rsp_ctx) == RSPCTX_PROCEEDING)
	    {
	      if (locked == 0)
		{
		  psp_core_lock ();
		  locked = 1;
		}
	      pspm_sfp_delete_broken_location (pspm, req);

	      if (0 == pspm_sfp_forward_200ok_final_response (pspm, req))
		{
		  pspm_sfp_cancel_all_pending_branches (pspm, req);
		}
	      else
		{
		  pspm_sfp_forward_provisionnal_response (pspm, req);

		  if (0 == psp_req_is_expired (req))
		    {
		      pspm_sfp_answer_request_and_cancel_branches (pspm, req,
								   408);
		    }
		  else if (req->locations == NULL
			   && 0 == pspm_all_branch_are_terminated (pspm, req)
			   && 0 == pspm_sfp_wait_for_final_response (pspm,
								     req))
		    {
		    }
		  else if (req->locations == NULL
			   && 0 == pspm_all_branch_are_terminated (pspm, req))
		    {
		      pspm_sfp_answer_request_and_cancel_branches (pspm, req,
								   503);
		    }
		}
	    }
	}

      /***********************************************************/
      /*    handle events in RSPCTX_PROCEEDING state             */
      for (pnext = pspm->psp_requests, req = pnext; pnext != NULL;
	   req = pnext)
	{
	  pnext = req->next;
	  if (rsp_ctx_get_state (req->rsp_ctx) == RSPCTX_COMPLETED
	      || rsp_ctx_get_state (req->rsp_ctx) == RSPCTX_TERMINATED)
	    {
	      if (locked == 0)
		{
		  psp_core_lock ();
		  locked = 1;
		}
	      pspm_sfp_delete_completed_branchs (pspm, req);
	      /* TODO?? pspm_sfp_forward_200ok (pspm, req); *//* ?? */
	    }
	}

      /***********************************************************/
      /*    handle events in RSPCTX_CANCELLED state             */
      for (pnext = pspm->psp_requests, req = pnext; pnext != NULL;
	   req = pnext)
	{
	  pnext = req->next;
	  if (rsp_ctx_get_state (req->rsp_ctx) == RSPCTX_CANCELLED)
	    {
	      if (locked == 0)
		{
		  psp_core_lock ();
		  locked = 1;
		}
	      /* pspm_sfp_reply_best_answer_or_404 (pspm, req); */
	      if (0 == pspm_sfp_forward_200ok_final_response (pspm, req))
		{		/* sent */
		}
	      else if (0 == pspm_sfp_wait_for_final_response (pspm, req))
		{
		}
	      else if (0 == psp_req_is_expired (req))
		pspm_sfp_answer_request_and_cancel_branches (pspm, req, 408);
	      pspm_sfp_delete_completed_branchs (pspm, req);
	    }
	}


      /***********************************************************/
      /*    handle events in RSPCTX_BROKEN state                 */
      for (pnext = pspm->psp_requests, req = pnext; pnext != NULL;
	   req = pnext)
	{
	  pnext = req->next;
	  if (rsp_ctx_get_state (req->rsp_ctx) == RSPCTX_BROKEN)
	    {
	      if (locked == 0)
		{
		  psp_core_lock ();
		  locked = 1;
		}
	      /* pspm_sfp_reply_best_answer_or_404 (pspm, req); */
	      pspm_sfp_delete_completed_branchs (pspm, req);
	    }
	}

      /* memory management for sfp transactions */
      if (locked == 0)
	{
	  psp_core_lock ();
	  locked = 1;
	}
      pspm_sfp_delete_broken_transactions (pspm);

      psp_core_unlock ();
      locked = 0;


      /* .... */
      /* pspm_sfp_unlock_psp_requests_access (pspm); */

      if (max == 0)
	return 0;
    }
}

int
pspm_sfp_cancel_pending_branch (pspm_sfp_t * pspm, psp_req_t * req,
				sfp_branch_t * sfp)
{
  transaction_t *transaction;
  sipevent_t *evt;
  route_t *route;
  sip_t *cancel;
  url_t *url;
  int i;

  if (sfp->out_tr->state == ICT_TERMINATED
      || sfp->out_tr->state == ICT_COMPLETED
      || sfp->out_tr->state == NICT_TERMINATED
      || sfp->out_tr->state == NICT_COMPLETED)
    return -1;

  if (sfp->out_tr->state == ICT_CALLING)
    return -1;

  i = osip_msg_build_cancel (&cancel, sfp->out_tr->orig_request);

  /* add the same route-set than in the previous request */
  url = NULL;
  if (!list_eol (sfp->out_tr->orig_request->routes, 0))
    {
      route = (route_t *) list_get (sfp->out_tr->orig_request->routes, 0);
      i = url_clone (route->url, &url);
      if (i != 0)
	{
	  return -1;
	}
    }

  evt = osip_new_outgoing_sipmessage (cancel);
  transaction = psp_core_uap_create_transaction (evt);
  if (transaction == NULL)
    {
      OSIP_TRACE (osip_trace
		  (__FILE__, __LINE__, OSIP_INFO3, NULL,
		   "sfp module: Could not create a CANCEL transaction!\n"));
      msg_free (evt->sip);
      sfree (evt->sip);
      sfree (evt);
      return -1;
    }
  evt->transactionid = transaction->transactionid;
  list_add (pspm->broken_transactions, transaction, 0);
  transaction_add_event (transaction, evt);
  return 0;
}

int
pspm_sfp_cancel_all_pending_branches (pspm_sfp_t * pspm, psp_req_t * req)
{
  /* 1: find remaining transaction to CANCEL,
     2: send a CANCEL */
  sfp_branch_t *sfp;

  for (sfp = req->rsp_ctx->branches; sfp != NULL; sfp = sfp->next)
    {
      if (sfp->out_tr != NULL)
	pspm_sfp_cancel_pending_branch (pspm, req, sfp);
    }
  return 0;
}

int
pspm_sfp_cancel_match_invite (transaction_t * invite, sip_t * cancel)
{
  generic_param_t *br;
  generic_param_t *br2;
  via_t *via;

  via_param_getbyname (invite->topvia, "branch", &br);
  via = list_get (cancel->vias, 0);
  if (via == NULL)
    return -1;			/* request without via??? */
  via_param_getbyname (via, "branch", &br2);
  if (br != NULL && br2 == NULL)
    return -1;
  if (br2 != NULL && br == NULL)
    return -1;
  if (br2 != NULL && br != NULL)	/* compliant UA  :) */
    {
      if (br->gvalue != NULL && br2->gvalue != NULL &&
	  0 == strcmp (br->gvalue, br2->gvalue))
	return 0;
      return -1;
    }
  /* old backward compatibility mechanism */
  if (0 != call_id_match (invite->callid, cancel->call_id))
    return -1;
  if (core->disable_check_for_to_tag_in_cancel == 0)
    {
      if (0 != to_tag_match (invite->to, cancel->to))
	return -1;
    }
  if (0 != from_tag_match (invite->from, cancel->from))
    return -1;
  if (0 != strcmp (invite->cseq->number, cancel->cseq->number))
    return -1;
  if (0 != via_match (invite->topvia, via))
    return -1;
  return 0;
}

int
pspm_sfp_handle_new_cancel (pspm_sfp_t * pspm, transaction_t * tr_cancel)
{
  psp_req_t *req;
  sipevent_t *evt;
  sip_t *response;
  int i;

  /* find the transaction */
  for (req = pspm->psp_requests; req != NULL; req = req->next)
    {
      if (0 ==
	  pspm_sfp_cancel_match_invite (req->inc_tr, tr_cancel->orig_request))
	{
	  break;
	}

    }

  list_add (pspm->broken_transactions, tr_cancel, 0);
  if (req == NULL
      || req->inc_tr->state == IST_COMPLETED
      || req->inc_tr->state == IST_TERMINATED)
    {
      if (req != NULL)
	{
	  OSIP_TRACE (osip_trace
		      (__FILE__, __LINE__, OSIP_INFO2, NULL,
		       "module sfp: Request already terminated.!\n"));
	}
      else
	{
	  OSIP_TRACE (osip_trace
		      (__FILE__, __LINE__, OSIP_WARNING, NULL,
		       "module sfp: No valid INVITE found; answer to request with 481.!\n"));
	}

      i =
	osip_msg_sfp_build_response (&response, 481, tr_cancel->orig_request);
      /* should we fix the to tag with one of the provisonnal response? */
      if (i != 0)
	{
	  OSIP_TRACE (osip_trace
		      (__FILE__, __LINE__, OSIP_ERROR, NULL,
		       "module sfp: Cannot build 481 response!\n"));
	  /* we got to cancel all requests */
	  rsp_ctx_set_state (req->rsp_ctx, RSPCTX_BROKEN);
	  return -1;
	}
      evt = osip_new_outgoing_sipmessage (response);
      evt->transactionid = tr_cancel->transactionid;
      transaction_add_event (tr_cancel, evt);
      psp_osip_wakeup (core->psp_osip);
      return 0;
    }


  OSIP_TRACE (osip_trace
	      (__FILE__, __LINE__, OSIP_INFO2, NULL,
	       "module sfp: Cancelling the request.!\n"));

  /* reply 408 Time out */
  i = osip_msg_sfp_build_response (&response, 200, tr_cancel->orig_request);
  /* should we fix the to tag with one of the provisonnal response? */
  if (i != 0)
    {
      OSIP_TRACE (osip_trace
		  (__FILE__, __LINE__, OSIP_ERROR, NULL,
		   "module sfp: Cannot build 200 response for CANCEL!\n"));
      /* we got to cancel all requests */
      rsp_ctx_set_state (req->rsp_ctx, RSPCTX_BROKEN);
      return -1;
    }
  evt = osip_new_outgoing_sipmessage (response);
  evt->transactionid = tr_cancel->transactionid;
  transaction_add_event (tr_cancel, evt);

  /* CANCEL all outgoing requests */
  pspm_sfp_cancel_all_pending_branches (pspm, req);

  rsp_ctx_set_state (req->rsp_ctx, RSPCTX_CANCELLED);
  psp_osip_wakeup (core->psp_osip);

  return 0;
}

int
pspm_all_branch_are_terminated (pspm_sfp_t * pspm, psp_req_t * req)
{
  sfp_branch_t *sfp;

  for (sfp = req->rsp_ctx->branches; sfp != NULL; sfp = sfp->next)
    {
      if (sfp->out_tr->state == ICT_TERMINATED
	  || sfp->out_tr->state == ICT_COMPLETED
	  || sfp->out_tr->state == NICT_TERMINATED
	  || sfp->out_tr->state == NICT_COMPLETED)
	{
	}
      else
	return -1;
    }
  return 0;
}

int
pspm_sfp_handle_ack (pspm_sfp_t * pspm, psp_req_t * req)
{
  location_t *ptr;
  location_t *loc;
  int i;


  /* This method should be called only when a route header exist
     in the request. (else ACK must be sent to the contact feild
     of 200 OK or are consumed by ACK > 299 */

  /* For statefull mode, the plugins have to provide
     locations to be resolved... */
  for (loc = req->locations; loc != NULL; loc = ptr)
    {
      int is_an_ip;

      ptr = loc->next;
      is_an_ip = inet_addr (loc->url->host);

      if (is_an_ip == -1)	/* not an ip address, ask for a resolution */
	{
	  psp_core_event_resolv_add_query (sgetcopy (loc->url->host));
	  OSIP_TRACE (osip_trace
		      (__FILE__, __LINE__, OSIP_ERROR, NULL,
		       "sfp module: we have to send the ACK now, but the ACK contains a FQDN!\n"));
	}
      else
	{
	  /* WE DON'T NEED TO WAIT! this location can be tried right now. */
	  sip_t *ack;
	  int port;

	  i = osip_msg_sfp_build_request_to_forward (req, &ack, loc->url,
						     loc->path,
						     req->request,
						     __IS_PSP_STAY_ON_PATH
						     (req->flag));
	  if (i != 0)
	    {
	      OSIP_TRACE (osip_trace
			  (__FILE__, __LINE__, OSIP_ERROR, NULL,
			   "sfp module: could not build ACK to forward!\n"));
	    }
	  else
	    {
	      i = pspm_sfp_call_plugins_for_request (pspm, req, NULL);
	      if (i != 0)
		{
		  OSIP_TRACE (osip_trace
			      (__FILE__, __LINE__, OSIP_ERROR, NULL,
			       "sfp module: Could not call plugins for this request!\n"));
		}
	    }
	  OSIP_TRACE (osip_trace
		      (__FILE__, __LINE__, OSIP_INFO2, NULL,
		       "sfp module: we have to send the ACK now!\n"));

	  if (loc->path!=NULL)
	    {
	      OSIP_TRACE (osip_trace
			  (__FILE__, __LINE__, OSIP_BUG, NULL,
			   "sfp module: We have to recalculate the route!\n"));
	      /* get the initial route */
	      /* TODO ... */
	    }

	  port = 5060;
	  if (loc->url->port != NULL)
	    port = satoi (loc->url->port);
	  psp_core_cb_snd_message (NULL, ack, loc->url->host, port, -1);
	  msg_free (ack);
	  sfree (ack);
	  loc->url = NULL;

	  REMOVE_ELEMENT (req->locations, loc);
	  location_free (loc);
	  sfree (loc);
	}
    }
  OSIP_TRACE (osip_trace
	      (__FILE__, __LINE__, OSIP_INFO2, NULL,
	       "sfp module: deleting ACK!\n"));
  psp_req_free (req);
  sfree (req);
  return 0;
}


int
pspm_sfp_create_sfp_branch (pspm_sfp_t * pspm, psp_req_t * req, url_t * url,
			    char *path, ppl_dns_ip_t * ips)
{
  transaction_t *transaction;
  sipevent_t *evt;
  sfp_branch_t *br;
  unsigned long int one_inet_addr;
  int i;
  url_t *url2;

  if (req->rsp_ctx == NULL)
    return -1;

  OSIP_TRACE (osip_trace
	      (__FILE__, __LINE__, OSIP_INFO3, NULL,
	       "sfp module: Creating new branch for statefull transaction!\n"));
  i = sfp_branch_init (&br, req->rsp_ctx->branch);
  if (i != 0)
    return -1;

  i = url_clone (url, &url2);
  if (i != 0)
    {
      sfp_branch_free (br);
      sfree (br);
      return -1;
    }

  /* if an ips->name is not an IP: we replace the url host with the FQDN.
     (In this case, we are sure ips->name is not a domain name but a real FQDN.) */
  if (ips != NULL && (int) (one_inet_addr = inet_addr (ips->name)) == -1)
    {
      OSIP_TRACE (osip_trace
		  (__FILE__, __LINE__, OSIP_INFO3, NULL,
		   "sfp module: Replacing with the FQDN (%s) for statefull transaction!\n",
		   ips->name));
      sfree (url2->host);
      url2->host = sgetcopy (ips->name);
    }

  sfp_branch_set_url (br, url2);

  if (ips != NULL && ips->next != NULL)	/* fallback exists */
    sfp_branch_clone_and_set_fallbacks (br, ips->next);

  i = url_clone (br->url, &url2);
  if (i != 0)
    {
      sfp_branch_free (br);
      sfree (br);
      return -1;
    }

  /* clone the request... start the new transaction... */
  i = osip_msg_sfp_build_request_to_forward (req, &(br->request), url2, path,
					     req->request,
					     __IS_PSP_STAY_ON_PATH (req->
								    flag));
  if (i != 0)
    {
      OSIP_TRACE (osip_trace
		  (__FILE__, __LINE__, OSIP_ERROR, NULL,
		   "sfp module: Could not create the request to forward for statefull transaction!\n"));
      sfp_branch_free (br);
      sfree (br);
      return -1;
    }
  i = pspm_sfp_call_plugins_for_request (pspm, req, br);
  if (i != 0)
    {
      OSIP_TRACE (osip_trace
		  (__FILE__, __LINE__, OSIP_ERROR, NULL,
		   "sfp module: Could not call plugins for this request!\n"));
      sfp_branch_free (br);
      sfree (br);
      return -1;
    }

  if (!MSG_IS_INVITE (br->request))
    /* I set the timeout to a lower value. I think it should be 160 for INVITE
       and 32 for other requests. I prefer to use lower value by now, but those
       values will be dynamic later... */
    sfp_branch_set_timeout (br, 32);

  /* create a new outgoing transaction and add the REQUEST */
  evt = osip_new_outgoing_sipmessage (br->request);
  br->request = NULL;		/* avoid double free and speed the proxy... */

  transaction = psp_core_uap_create_transaction (evt);
  if (transaction == NULL)
    {
      OSIP_TRACE (osip_trace
		  (__FILE__, __LINE__, OSIP_ERROR, NULL,
		   "sfp module: Could not create a transaction for this request!\n"));
      msg_free(evt->sip);
      sfree(evt->sip);
      sfree(evt);
      sfp_branch_free (br);
      sfree (br);
      return -1;
    }
  sfp_branch_set_transaction (br, transaction);
  /* The oSIP stack transaction element has an entry for the resolved IP
     address where the stack should send the request. As we are using
     probably using a FQDN in the rq-uri (or the route), we can force
     the stack to keep the FQDN value there while it still can access
     the real IP of this host. */
  if (ips != NULL)
    {
      if (ips->srv_ns_flag == PSP_SRV_LOOKUP)	/* take the resolved port */
	{
	}
      else			/* for NS_LOOKUP, always rewrite the port number of the transaction */
	{
	  if (br->url->port != NULL)
	    ips->sin.sin_port = htons ((short) satoi (br->url->port));
	}

      if (MSG_IS_INVITE (req->request))
	ict_set_destination (transaction->ict_context,
			     sgetcopy (inet_ntoa (ips->sin.sin_addr)),
			     ntohs (ips->sin.sin_port));
      else
	nict_set_destination (transaction->nict_context,
			      sgetcopy (inet_ntoa (ips->sin.sin_addr)),
			      ntohs (ips->sin.sin_port));
      OSIP_TRACE (osip_trace
		  (__FILE__, __LINE__, OSIP_INFO2, NULL,
		   "sfp module: The destination for the transaction is set to: '%s'.!\n",
		   inet_ntoa (ips->sin.sin_addr)));
    }

#if defined(ALPHA_FIREWALL)
  /* the firewall is based on the call-id value of the call. We should
     use at least callid+to+from instead. */

  if (core->nat_enabled == 1 && MSG_IS_BYE(evt->sip))
    {
      char *hash_value;
      i = call_id_2char(evt->sip->call_id, &hash_value);
      if (i == 0)
	{
	  firewall_entry_remove_rule (hash_value);
	  sfree(hash_value);
	}
      /* continue anyway. */
    }
  else if (core->nat_enabled == 1 && MSG_IS_INVITE(evt->sip))
    {
      char *host;
      int port;
      transaction_get_destination(transaction, &host, &port);
      if (host==NULL) /* In this case, the message is sent to this url */
	host = evt->sip->strtline->rquri->host;

      i = -1;
      if (host!=NULL && inet_addr (host) != -1)
	{
	  i = firewall_fix_invite_to_forward (evt->sip, host);
	}
      else
	{
	  OSIP_TRACE (osip_trace
		      (__FILE__, __LINE__, OSIP_ERROR, NULL,
		       "module sfp: firewall module failed (no target_host defined)!\n"));
	  list_add (pspm->broken_transactions, br->out_tr, 0);
	  br->out_tr = NULL;
	  msg_free(evt->sip);
	  sfree(evt->sip);
	  sfree(evt);
	  sfp_branch_free (br);
	  sfree (br);
	  return -1;
	}

      if (i == -1)
	{
	  OSIP_TRACE (osip_trace
		      (__FILE__, __LINE__, OSIP_ERROR, NULL,
		       "module sfp: firewall module failed!\n"));
	  list_add (pspm->broken_transactions, br->out_tr, 0);
	  br->out_tr = NULL;
	  msg_free(evt->sip);
	  sfree(evt->sip);
	  sfree(evt);
	  sfp_branch_free (br);
	  sfree (br);
	  return -1;
	}
      else if (i == 0)
	{
	  OSIP_TRACE (osip_trace
		      (__FILE__, __LINE__, OSIP_INFO2, NULL,
		       "module sfp: firewall module operation done!\n"));
	}
      else
	{
	  OSIP_TRACE (osip_trace
		      (__FILE__, __LINE__, OSIP_INFO2, NULL,
		       "module sfp: No firewall module operation done!\n"));
	}
    }
#endif

  /* else, the usual processing is done */
  evt->transactionid = transaction->transactionid;
  transaction_add_event (transaction, evt);

  ADD_ELEMENT (req->rsp_ctx->branches, br);

  OSIP_TRACE (osip_trace
	      (__FILE__, __LINE__, OSIP_INFO2, NULL,
	       "sfp module: transaction correctly forwarded!\n"));

  rsp_ctx_set_state (req->rsp_ctx, RSPCTX_PROCEEDING);

  psp_osip_wakeup (core->psp_osip);
  return 0;
}

int
pspm_sfp_handle_new_request (pspm_sfp_t * pspm, psp_req_t * req)
{
  location_t *loc;
  location_t *ptr;
  int i;

  /* refuse plugin who gives non sip/sips-urls. */
  for (loc = req->locations; loc != NULL; loc = ptr)
    {
      ptr = loc->next;
      if (0 != strcmp (loc->url->scheme, "sip")
	  && 0 != strcmp (loc->url->scheme, "sips"))
	{
	  OSIP_TRACE (osip_trace
		      (__FILE__, __LINE__, OSIP_WARNING, NULL,
		       "sfp module: Location should be a sip url!\n"));
	  REMOVE_ELEMENT (req->locations, loc);
	  location_free (loc);
	  sfree (loc);
	}
    }

  /* In case of record-routing, the proxy will allocate the location
     itself. */
  if (req->locations == NULL)
    {
      route_t *route;
      route_t *route2;
      url_t *dest_url;
      url_t *url;
      msg_getroute (req->request, 0, &route);
      msg_getroute (req->request, 1, &route2);
      if (route == NULL)
	{
	  /* ??? */
	  OSIP_TRACE (osip_trace
		      (__FILE__, __LINE__, OSIP_ERROR, NULL,
		       "sfp module: no valid location given?\n"));
	  list_add (pspm->broken_transactions, req->inc_tr, 0);
	  req->inc_tr = NULL;
	  psp_req_free (req);
	  sfree (req);
	  return -1;
	}
      else if (0 != psp_core_is_responsible_for_this_route (route->url))
	/* use the request-uri */
	dest_url = route->url;
      else if (route2 != NULL)
	/* use the route header #2 */
	dest_url = route2->url;
      else
	/* use the request-uri */
	dest_url = req->request->strtline->rquri;

      i = url_clone (dest_url, &url);
      if (i != 0)
	{
	  OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_ERROR, NULL,
				  "sfp module: Could not clone request-uri!\n"));
	  list_add (pspm->broken_transactions, req->inc_tr, 0);
	  req->inc_tr = NULL;
	  psp_req_free (req);
	  sfree (req);
	  return -1;
	}
      i = location_init (&loc, url, 3600);
      if (i != 0)
	{
	  url_free (url);
	  sfree (url);
	  OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_BUG, NULL,
				  "sfp module: Could not create location info!\n"));
	  list_add (pspm->broken_transactions, req->inc_tr, 0);
	  req->inc_tr = NULL;
	  psp_req_free (req);
	  sfree (req);
	  return -1;
	}
      ADD_ELEMENT (req->locations, loc);
      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO2, NULL,
			      "sfp module: location guessed from route headers!\n"));
    }

  if (MSG_IS_ACK (req->request))
    return pspm_sfp_handle_ack (pspm, req);

  /* create a sfull context */
  if (req->rsp_ctx == NULL)
    {
      i = psp_core_sfp_new_rsp_ctx (&(req->rsp_ctx), req);
      if (i != 0)
	{
	  req->rsp_ctx = NULL;
	  OSIP_TRACE (osip_trace
		      (__FILE__, __LINE__, OSIP_ERROR, NULL,
		       "sfp module: failed to create a rsp_ctx for this incoming transaction!\n"));
	  return -1;
	}
    }

  /* Initially, the incoming transaction is in the RSPCTX_CALLING state */

  OSIP_TRACE (osip_trace
	      (__FILE__, __LINE__, OSIP_INFO3, NULL,
	       "sfp module: Request added in psp_requests!\n"));
  ADD_ELEMENT (pspm->psp_requests, req);

  /* For statefull mode, the plugins have to provide
     locations to be resolved... */

  for (loc = req->locations; loc != NULL; loc = ptr)
    {
      int is_an_ip;
      ptr = loc->next;
      is_an_ip = inet_addr (loc->url->host);

      if (is_an_ip == -1)	/* not an ip address, ask for a resolution */
	{
	  psp_core_event_resolv_add_query (sgetcopy (loc->url->host));
	}
      else
	{
	  /* WE DON'T NEED TO WAIT! this location can be tried right now. */
	  i = pspm_sfp_create_sfp_branch (pspm, req, loc->url, loc->path, NULL);
	  /* in all cases, (failure or success), this location
	     has been tried and will never be tried again */
	  REMOVE_ELEMENT (req->locations, loc);
	  location_free (loc);
	  sfree (loc);
	}
    }

  return 0;
}

int
pspm_sfp_handle_new_location (pspm_sfp_t * pspm, psp_req_t * req)
{
  location_t *loc;
  location_t *ptr;
  int i;

  /* this method is regularly called and check if some pending
     statefull transactions have new locations to try. */

  for (loc = req->locations; loc != NULL; loc = ptr)
    {
      int is_an_ip;

      ptr = loc->next;
      is_an_ip = inet_addr (loc->url->host);

      if (is_an_ip == -1)
	{
	  ppl_dns_entry_t *dns_result;

	  i = ppl_dns_get_result (&dns_result, loc->url->host);
	  if (i != 0)
	    {
	      OSIP_TRACE (osip_trace
			  (__FILE__, __LINE__, OSIP_INFO3, NULL,
			   "module sfp: Keep waiting for the resolution to succeed...!\n"));
	    }
	  else
	    {
	      /* The resolver can provide several IPs for the same location.
	         This is a major issue here: What should I do: starting a
	         transactions for each IP does not seems to be reasonnable.
	         (forking should be done on several locations, not
	         on one location that resolved to many IPs. In the case
	         where a host is resolved to many IPs, I sould try the
	         first one and use the other ones as fallbacks.
	       */
	      i = pspm_sfp_create_sfp_branch (pspm, req, loc->url, loc->path,
					      dns_result->dns_ips);
	      /* in all cases, (failure or success), this location
	         has been tried and will never be tried again */
	      REMOVE_ELEMENT (req->locations, loc);
	      location_free (loc);
	      sfree (loc);
	    }
	}
      else
	{
	  OSIP_TRACE (osip_trace
		      (__FILE__, __LINE__, OSIP_BUG, NULL,
		       "module sfp: An ip address is found here! (I don't expect it here)!\n"));
	}
    }

  return 0;
}

int
pspm_sfp_delete_broken_location (pspm_sfp_t * pspm, psp_req_t * req)
{
  location_t *loc;
  location_t *ptr;

  for (loc = req->locations; loc != NULL; loc = ptr)
    {
      ppl_dns_error_t *error;

      ptr = loc->next;
      if (0 == ppl_dns_get_error (&error, loc->url->host))
	{

	  OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO2, NULL,
				  "sfp module: one bad location is removed!\n"));
	  REMOVE_ELEMENT (req->locations, loc);
	  location_free (loc);
	  sfree (loc);
	}
    }
  return 0;
}

int
pspm_sfp_delete_broken_branchs (pspm_sfp_t * pspm, psp_req_t * req)
{

  /* TODO... */

  return 0;
}


int
pspm_sfp_delete_completed_branchs (pspm_sfp_t * pspm, psp_req_t * req)
{
  sfp_branch_t *br;
  sfp_branch_t *ptr;

  if (req->inc_tr == NULL)
    {
    }
  else if (req->inc_tr->state != IST_TERMINATED
	   && req->inc_tr->state != NIST_TERMINATED)
    return -1;

  if (req->rsp_ctx == NULL)
    {
      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO2, NULL,
			      "sfp module: pspm_sfp_delete_completed_branchs delete this psp_req!\n"));
      REMOVE_ELEMENT (pspm->psp_requests, req);
      psp_req_free (req);
      sfree (req);
      return 0;
    }

  ptr = NULL;
  for (br = req->rsp_ctx->branches; br != NULL; br = br->next)
    {
      if (br->out_tr == NULL)
	{
	}
      else if (br->out_tr->state != ICT_TERMINATED
	       && br->out_tr->state != NICT_TERMINATED)
	return -1;
    }

  OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO2, NULL,
			  "sfp module: pspm_sfp_delete_completed_branchs delete this psp_req!\n"));
  REMOVE_ELEMENT (pspm->psp_requests, req);
  psp_req_free (req);
  sfree (req);
  return 0;
}


int
pspm_sfp_answer_request_and_cancel_branches (pspm_sfp_t * pspm,
					     psp_req_t * req, int code)
{
  sip_t *response;
  sipevent_t *evt;
  int i;

  OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO2, NULL,
			  "sfp module: answer request with code: %i!\n",
			  code));

  pspm_sfp_cancel_all_pending_branches (pspm, req);
  psp_osip_wakeup (core->psp_osip);

  i =
    osip_msg_sfp_build_response (&response, code, req->inc_tr->orig_request);
  if (i != 0)
    {
      OSIP_TRACE (osip_trace
		  (__FILE__, __LINE__, OSIP_ERROR, NULL,
		   "module sfp: Cannot build 408 response!\n"));
      rsp_ctx_set_state (req->rsp_ctx, RSPCTX_BROKEN);
      return -2;
    }
  evt = osip_new_outgoing_sipmessage (response);
  evt->transactionid = req->inc_tr->transactionid;

  OSIP_TRACE (osip_trace
	      (__FILE__, __LINE__, OSIP_ERROR, NULL,
	       "module sfp: No valid location found; answer to request with 480.!\n"));
  req->response = response;
  i = pspm_sfp_call_plugins_for_snd_response (pspm, req);
  if (i != 0)
    {
      OSIP_TRACE (osip_trace
		  (__FILE__, __LINE__, OSIP_ERROR, NULL,
		   "module sfp: Can't call plugins for response.!\n"));
      rsp_ctx_set_state (req->rsp_ctx, RSPCTX_BROKEN);
      req->response = NULL;
      return -2;
    }
  req->response = NULL;
  transaction_add_event (req->inc_tr, evt);
  rsp_ctx_set_state (req->rsp_ctx, RSPCTX_BROKEN);
  psp_osip_wakeup (core->psp_osip);
  return 0;
}

int
pspm_sfp_wait_for_final_response (pspm_sfp_t * pspm, psp_req_t * req)
{
  sfp_branch_t *br;
  sfp_branch_t *ptr;
  sip_t *response;
  sipevent_t *evt;
  int i;

  OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO2, NULL,
			  "sfp module: Waiting for a final answer!\n"));


  /* send a 408 if no final answer has been received */
  ptr = NULL;
  for (br = req->rsp_ctx->branches; br != NULL; br = br->next)
    {
      if (br->out_tr->state == ICT_TERMINATED
	  || br->out_tr->state == ICT_COMPLETED
	  || br->out_tr->state == NICT_TERMINATED
	  || br->out_tr->state == NICT_COMPLETED)
	{
	  if (ptr == NULL)
	    ptr = br;
	  else if (0 <
		   strcmp (ptr->out_tr->last_response->strtline->statuscode,
			   br->out_tr->last_response->strtline->statuscode))
	    ptr = br;		/* better response */
	}
    }

  if (ptr == NULL)
    return -1;			/* no final response yet */

  /* at least one final answer has been received. It is
     simple if we only get one! but could be tricky if more than
     one has been received...
     From rfc3261: Section 16.7 point 6:

     6.  Choosing the best response

     A stateful proxy MUST send a final response to a response
     context's server transaction if no final responses have been
     immediately forwarded by the above rules and all client
     transactions in this response context have been terminated.

     The stateful proxy MUST choose the "best" final response among
     those received and stored in the response context.

     If there are no final responses in the context, the proxy MUST
     send a 408 (Request Timeout) response to the server
     transaction.

     Otherwise, the proxy MUST forward a response from the responses
     stored in the response context.  It MUST choose from the 6xx
     class responses if any exist in the context.  If no 6xx class
     responses are present, the proxy SHOULD choose from the lowest
     response class stored in the response context.  The proxy MAY
     select any response within that chosen class.  The proxy SHOULD
     give preference to responses that provide information affecting
     resubmission of this request, such as 401, 407, 415, 420, and
     484 if the 4xx class is chosen.

     A proxy which receives a 503 (Service Unavailable) response
     SHOULD NOT forward it upstream unless it can determine that any
     subsequent requests it might proxy will also generate a 503.
     In other words, forwarding a 503 means that the proxy knows it
     cannot service any requests, not just the one for the Request-
     URI in the request which generated the 503.  If the only
     response that was received is a 503, the proxy SHOULD generate
     a 500 response and forward that upstream.

     The forwarded response MUST be processed as described in steps
     "Aggregate Authorization Header Field Values" through "Record-
     Route".

     For example, if a proxy forwarded a request to 4 locations, and
     received 503, 407, 501, and 404 responses, it may choose to
     forward the 407 (Proxy Authentication Required) response.

     1xx and 2xx responses may be involved in the establishment of
     dialogs.  When a request does not contain a To tag, the To tag
     in the response is used by the UAC to distinguish multiple
     responses to a dialog creating request.  A proxy MUST NOT
     insert a tag into the To header field of a 1xx or 2xx response
     if the request did not contain one.  A proxy MUST NOT modify
     the tag in the To header field of a 1xx or 2xx response.

     Since a proxy may not insert a tag into the To header field of
     a 1xx response to a request that did not contain one, it cannot
     issue non-100 provisional responses on its own.  However, it
     can branch the request to a UAS sharing the same element as the
     proxy.  This UAS can return its own provisional responses,
     entering into an early dialog with the initiator of the
     request.  The UAS does not have to be a discreet process from
     the proxy.  It could be a virtual UAS implemented in the same
     code space as the proxy.

     3-6xx responses are delivered hop-by-hop.  When issuing a 3-6xx
     response, the element is effectively acting as a UAS, issuing
     its own response, usually based on the responses received from
     downstream elements.  An element SHOULD preserve the To tag
     when simply forwarding a 3-6xx response to a request that did
     not contain a To tag.

     A proxy MUST NOT modify the To tag in any forwarded response to
     a request that contains a To tag.

     While it makes no difference to the upstream elements if the
     proxy replaced the To tag in a forwarded 3-6xx response,
     preserving the original tag may assist with debugging.

     When the proxy is aggregating information from several
     responses, choosing a To tag from among them is arbitrary, and
     generating a new To tag may make debugging easier.  This
     happens, for instance, when combining 401 (Unauthorized) and
     407 (Proxy Authentication Required) challenges, or combining
     Contact values from unencrypted and unauthenticated 3xx
     responses.

     TODO!

     ouch.... :/
   */

  OSIP_TRACE (osip_trace
	      (__FILE__, __LINE__, OSIP_INFO3, NULL,
	       "module sfp: We have a xxx>299 response, let's try to forward it!\n"));
  /* we have a xxx final answer! forward the best... */
  i =
    osip_msg_sfp_build_response_to_forward (&response,
					    ptr->out_tr->last_response);

  if (i != 0)
    {
      OSIP_TRACE (osip_trace
		  (__FILE__, __LINE__, OSIP_ERROR, NULL,
		   "module sfp: Cannot forward xxx response!\n"));
      rsp_ctx_set_state (req->rsp_ctx, RSPCTX_BROKEN);
      return 0;
    }

  req->response = response;
  i = pspm_sfp_call_plugins_for_snd_response (pspm, req);
  if (i != 0)
    {
      OSIP_TRACE (osip_trace
		  (__FILE__, __LINE__, OSIP_ERROR, NULL,
		   "module sfp: Could not call plugins!\n"));
      msg_free (response);
      sfree (response);
      req->response = NULL;
      rsp_ctx_set_state (req->rsp_ctx, RSPCTX_BROKEN);
      return 0;
    }
  req->response = NULL;
  evt = osip_new_outgoing_sipmessage (response);
  evt->transactionid = req->inc_tr->transactionid;
  transaction_add_event (req->inc_tr, evt);
  psp_osip_wakeup (core->psp_osip);

  rsp_ctx_set_state (req->rsp_ctx, RSPCTX_TERMINATED);

  OSIP_TRACE (osip_trace
	      (__FILE__, __LINE__, OSIP_INFO2, NULL,
	       "module sfp: Response xxx forwarded!\n"));
  return 0;
}

int
pspm_sfp_delete_broken_transactions (pspm_sfp_t * pspm)
{
  transaction_t *tr;
  int pos = 0;

  while (!list_eol (pspm->broken_transactions, pos))
    {
      tr = list_get (pspm->broken_transactions, pos);
      if (tr->state == ICT_TERMINATED
	  || tr->state == NICT_TERMINATED
	  || tr->state == IST_TERMINATED || tr->state == NIST_TERMINATED)
	{
	  list_remove (pspm->broken_transactions, pos);
	  transaction_free2 (tr);
	  sfree (tr);
	}
      else
	{
	  /* case where the transaction is more than 200s (broken) */
	  int now = ppl_time ();

	  if (now - tr->birth_time > 200)
	    {
	      list_remove (pspm->broken_transactions, pos);
	      transaction_free (tr);
	      sfree (tr);
	    }
	  else
	    pos++;
	}
    }
  return 0;
}

int
pspm_sfp_forward_response (pspm_sfp_t * pspm, psp_req_t * req,
			   sfp_branch_t * sfp)
{
  sipevent_t *evt;
  sip_t *responsexxx;
  int i;

  OSIP_TRACE (osip_trace
	      (__FILE__, __LINE__, OSIP_INFO3, NULL,
	       "module sfp: We have a xxx>200 response, let's try to forward it!\n"));

  i = osip_msg_sfp_build_response_to_forward (&responsexxx,
					      sfp->out_tr->last_response);
  if (i != 0)
    {
      OSIP_TRACE (osip_trace
		  (__FILE__, __LINE__, OSIP_ERROR, NULL,
		   "module sfp: Cannot forward xxx response!\n"));
      return -1;
    }

  req->response = responsexxx;

#if defined(ALPHA_FIREWALL)
  if (core->nat_enabled == 1)
    {
      i =
	firewall_fix_200ok_to_forward (responsexxx);
      if (i == -1)
	{
	  OSIP_TRACE (osip_trace
		      (__FILE__, __LINE__, OSIP_ERROR, NULL,
		       "module sfp: firewall module failed!\n"));
	  rsp_ctx_set_state (req->rsp_ctx, RSPCTX_BROKEN);
	  return 0;
	}
      else if (i == 0)
	{
	  OSIP_TRACE (osip_trace
		      (__FILE__, __LINE__, OSIP_INFO2, NULL,
		       "module sfp: firewall module operation done!\n"));
	}
      else
	{
	  OSIP_TRACE (osip_trace
		      (__FILE__, __LINE__, OSIP_INFO2, NULL,
		       "module sfp: No firewall module operation done!\n"));
	}
    }
#endif

  i = pspm_sfp_call_plugins_for_snd_response (pspm, req);
  if (i != 0)
    {
      OSIP_TRACE (osip_trace
		  (__FILE__, __LINE__, OSIP_ERROR, NULL,
		   "module sfp: Could not call plugins!\n"));
      msg_free (req->response);
      sfree (req->response);
      req->response = NULL;
      return -1;
    }

  evt = osip_new_outgoing_sipmessage (responsexxx);
  evt->transactionid = req->inc_tr->transactionid;
  i = psp_core_find_transaction_and_add_event (evt);
  if (i != 0)
    {
      OSIP_TRACE (osip_trace
		  (__FILE__, __LINE__, OSIP_ERROR, NULL,
		   "module sfp: Cannot find the transaction for the event!\n"));
      msg_free (evt->sip);
      sfree (evt->sip);
      sfree (evt);
      req->response = NULL;
      return -1;
    }

  OSIP_TRACE (osip_trace
	      (__FILE__, __LINE__, OSIP_INFO3, NULL,
	       "module sfp: Response xxx forwarded!\n"));

  sfp_branch_set_last_code_forwarded (sfp,
				      satoi (sfp->out_tr->last_response->
					     strtline->statuscode));
  return 0;
}

int
pspm_sfp_forward_200ok_final_response (pspm_sfp_t * pspm, psp_req_t * req)
{
  sfp_branch_t *sfpnext;
  sfp_branch_t *sfp;
  int done = 0;

  /* if a 200 OK has been received, we can end this transaction */
  for (sfpnext = req->rsp_ctx->branches, sfp = sfpnext;
       sfpnext != NULL; sfp = sfpnext)
    {
      sfpnext = sfp->next;
      if (sfp->out_tr->state == ICT_TERMINATED
	  || sfp->out_tr->state == NICT_TERMINATED)
	{
	  /* forward Message (if it's a new one...) */
	  if (sfp->out_tr->last_response != NULL
	      && sfp->out_tr->last_response->strtline != NULL
	      && sfp->out_tr->last_response->strtline->statuscode != NULL
	      && satoi (sfp->out_tr->last_response->strtline->statuscode) ==
	      200)
	    {
	      if (-1 == pspm_sfp_forward_response (pspm, req, sfp))
		{
		  OSIP_TRACE (osip_trace
			      (__FILE__, __LINE__, OSIP_ERROR, NULL,
			       "module sfp: pspm_sfp_forward_200ok_final_response could not forward response!\n"));
		  list_add (pspm->broken_transactions, sfp->out_tr, 0);
		  sfp->out_tr = NULL;
		  REMOVE_ELEMENT (req->rsp_ctx->branches, sfp);
		  sfp_branch_free (sfp);
		  sfree (sfp);
		}
	      else
		{
		  sfp_branch_set_last_code_forwarded (sfp,
						      satoi (sfp->out_tr->
							     last_response->
							     strtline->
							     statuscode));
		  OSIP_TRACE (osip_trace
			      (__FILE__, __LINE__, OSIP_INFO2, NULL,
			       "module sfp: pspm_sfp_forward_200ok_final_response!\n"));
		  rsp_ctx_set_state (req->rsp_ctx, RSPCTX_TERMINATED);
		  done = 1;
		}
	    }
	}
    }

  if (done == 1)
    return 0;
  return -1;
}

int
pspm_sfp_forward_provisionnal_response (pspm_sfp_t * pspm, psp_req_t * req)
{
  sfp_branch_t *sfpnext;
  sfp_branch_t *sfp;

  /* if a provisionnal response has been received, we can forward it. */
  for (sfpnext = req->rsp_ctx->branches, sfp = sfpnext;
       sfpnext != NULL; sfp = sfpnext)
    {
      sfpnext = sfp->next;
      if (sfp->out_tr->state != ICT_TERMINATED
	  && sfp->out_tr->state != NICT_TERMINATED
	  && sfp->out_tr->state != ICT_COMPLETED
	  && sfp->out_tr->state != NICT_COMPLETED)
	{
	  /* forward Message (if it's a new one...) */
	  if (sfp->out_tr->last_response != NULL
	      && sfp->out_tr->last_response->strtline != NULL
	      && sfp->out_tr->last_response->strtline->statuscode != NULL
	      && satoi (sfp->out_tr->last_response->strtline->statuscode) >
	      sfp_branch_get_last_code_forwarded (sfp))
	    {
	      if (-1 == pspm_sfp_forward_response (pspm, req, sfp))
		{
		  list_add (pspm->broken_transactions, sfp->out_tr, 0);
		  sfp->out_tr = NULL;
		  REMOVE_ELEMENT (req->rsp_ctx->branches, sfp);
		  sfp_branch_free (sfp);
		  sfree (sfp);
		}
	      else
		{
		  sfp_branch_set_last_code_forwarded (sfp,
						      satoi (sfp->out_tr->
							     last_response->
							     strtline->
							     statuscode));
		  OSIP_TRACE (osip_trace
			      (__FILE__, __LINE__, OSIP_INFO2, NULL,
			       "module sfp: pspm_sfp_forward_provisionnal_final_response!\n"));
		}

	    }
	}
    }

  return 0;
}

int
pspm_sfp_call_plugins_for_request (pspm_sfp_t * pspm, psp_req_t * req,
				   sfp_branch_t * sfp_branch)
{
  sfp_fwd_func_tab_t *t;
  sfp_fwd_func_t *f;

  if (MSG_IS_INVITE (req->request))
    t = pspm->fwd_invites;
  else if (MSG_IS_ACK (req->request))
    t = pspm->fwd_acks;
  else if (MSG_IS_REGISTER (req->request))
    t = pspm->fwd_registers;
  else if (MSG_IS_BYE (req->request))
    t = pspm->fwd_byes;
  else if (MSG_IS_OPTIONS (req->request))
    t = pspm->fwd_optionss;
  else if (MSG_IS_INFO (req->request))
    t = pspm->fwd_infos;
  else if (MSG_IS_CANCEL (req->request))
    t = pspm->fwd_cancels;
  else if (MSG_IS_NOTIFY (req->request))
    t = pspm->fwd_notifys;
  else if (MSG_IS_SUBSCRIBE (req->request))
    t = pspm->fwd_subscribes;
  else
    t = pspm->fwd_unknowns;

  for (f = t->func_hook_really_first; f != NULL; f = f->next)
    {
      f->cb_fwd_request_func (req);
      if (__IS_PSP_STOP (req->flag))
	return 0;
    }
  for (f = t->func_hook_first; f != NULL; f = f->next)
    {
      f->cb_fwd_request_func (req);
      if (__IS_PSP_STOP (req->flag))
	return 0;
    }
  for (f = t->func_hook_middle; f != NULL; f = f->next)
    {
      f->cb_fwd_request_func (req);
      if (__IS_PSP_STOP (req->flag))
	return 0;
    }
  for (f = t->func_hook_last; f != NULL; f = f->next)
    {
      f->cb_fwd_request_func (req);
      if (__IS_PSP_STOP (req->flag))
	return 0;
    }
  for (f = t->func_hook_really_last; f != NULL; f = f->next)
    {
      f->cb_fwd_request_func (req);
      if (__IS_PSP_STOP (req->flag))
	return 0;
    }
  return 0;
}

int
pspm_sfp_call_plugins_for_rcv_response (pspm_sfp_t * pspm, psp_req_t * req,
					sfp_branch_t * branch)
{
  sfp_rcv_func_tab_t *t;
  sfp_rcv_func_t *f;

  if (MSG_IS_STATUS_1XX (branch->response))
    t = pspm->rcv_1xxs;
  else if (MSG_IS_STATUS_2XX (branch->response))
    t = pspm->rcv_2xxs;
  else if (MSG_IS_STATUS_3XX (branch->response))
    t = pspm->rcv_3xxs;
  else if (MSG_IS_STATUS_4XX (branch->response))
    t = pspm->rcv_4xxs;
  else if (MSG_IS_STATUS_5XX (branch->response))
    t = pspm->rcv_5xxs;
  else if (MSG_IS_STATUS_6XX (branch->response))
    t = pspm->rcv_6xxs;
  else
    {
      printf ("ERROR: I do not understand this response");
      return -1;
    }

  for (f = t->func_hook_really_first; f != NULL; f = f->next)
    {
      f->cb_rcv_answer_func (req);
      if (__IS_PSP_STOP (req->flag))
	return 0;
    }
  for (f = t->func_hook_first; f != NULL; f = f->next)
    {
      f->cb_rcv_answer_func (req);
      if (__IS_PSP_STOP (req->flag))
	return 0;
    }
  for (f = t->func_hook_middle; f != NULL; f = f->next)
    {
      f->cb_rcv_answer_func (req);
      if (__IS_PSP_STOP (req->flag))
	return 0;
    }
  for (f = t->func_hook_last; f != NULL; f = f->next)
    {
      f->cb_rcv_answer_func (req);
      if (__IS_PSP_STOP (req->flag))
	return 0;
    }
  for (f = t->func_hook_really_last; f != NULL; f = f->next)
    {
      f->cb_rcv_answer_func (req);
      if (__IS_PSP_STOP (req->flag))
	return 0;
    }
  return 0;
}


int
pspm_sfp_call_plugins_for_snd_response (pspm_sfp_t * pspm, psp_req_t * req)
{
  sfp_snd_func_tab_t *t;
  sfp_snd_func_t *f;

  if (MSG_IS_STATUS_1XX (req->response))
    t = pspm->snd_1xxs;
  else if (MSG_IS_STATUS_2XX (req->response))
    t = pspm->snd_2xxs;
  else if (MSG_IS_STATUS_3XX (req->response))
    t = pspm->snd_3xxs;
  else if (MSG_IS_STATUS_4XX (req->response))
    t = pspm->snd_4xxs;
  else if (MSG_IS_STATUS_5XX (req->response))
    t = pspm->snd_5xxs;
  else if (MSG_IS_STATUS_6XX (req->response))
    t = pspm->snd_6xxs;
  else
    {
      printf ("ERROR: I do not understand this response");
      return -1;
    }

  for (f = t->func_hook_really_first; f != NULL; f = f->next)
    {
      f->cb_snd_answer_func (req);
      if (__IS_PSP_STOP (req->flag))
	return 0;
    }
  for (f = t->func_hook_first; f != NULL; f = f->next)
    {
      f->cb_snd_answer_func (req);
      if (__IS_PSP_STOP (req->flag))
	return 0;
    }
  for (f = t->func_hook_middle; f != NULL; f = f->next)
    {
      f->cb_snd_answer_func (req);
      if (__IS_PSP_STOP (req->flag))
	return 0;
    }
  for (f = t->func_hook_last; f != NULL; f = f->next)
    {
      f->cb_snd_answer_func (req);
      if (__IS_PSP_STOP (req->flag))
	return 0;
    }
  for (f = t->func_hook_really_last; f != NULL; f = f->next)
    {
      f->cb_snd_answer_func (req);
      if (__IS_PSP_STOP (req->flag))
	return 0;
    }
  return 0;
}

